/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/frame.c - 27/06/1999
 *
 * 
*/

#include <helium.h>

/* Type et union du package 
*/
#define  OBJ_TYPE	Obj_frame
#define  OBJ_UNION	obj.frame


static He_method he_meth_frame;

static char *he_option_geom = NULL;
static int he_iconifying = FALSE,
           he_wmdeco_test = FALSE,
           he_wmdeco_xa = 0, he_wmdeco_ya = 0,
           he_wmdeco_xb = 0, he_wmdeco_yb = 0;
           
/* 
 * Variables globales et exportees
*/
int he_wmdeco_width = 0, he_wmdeco_height = 0;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 *  Cration d'un frame.
 *  Renvoie le noeud ou NULL.
*/

He_node *HeCreateFrame ()
{
    He_node *hn;
    OBJ_TYPE *ob;
    
    HeFrameAutoInit();

    /* On cree le noeud */
    hn = HeBaseCreate (NULL, &he_meth_frame);
    if (hn == NULL) return NULL;

    /* On cree l'objet */
    ob = hn->OBJ_UNION = malloc (sizeof (OBJ_TYPE));
    if (ob == NULL) {
	HeError ("HeFrameCreate: malloc error\n");
	return NULL;
    }
    
    /* Init des champs du noeud */
    HE_BIT_CLR (hn->status, HE_STA_SHOW);
    hn->xm = 400;
    hn->ym = 300;

    /* Init des champs de l'objet */
    ob->close_proc = NULL;
    ob->resize_proc = NULL;
    ob->replace_proc = NULL;
    
    /* pour detecter les replace et resize 
       ici on force un replace et un resize  la creation. */
    ob->old_xb = ob->old_yb = ob->old_xm = ob->old_ym = -1;    
    
    HeFocusInitInfo (&ob->focus_info);

    /* Creation du window avec Xlib */
    hn->win = XCreateSimpleWindow (he_display, he_root_window,
	hn->xb, hn->yb, hn->xm, hn->ym, 
	hn->win_border,
	he_black,
	he_col_bg);

    /* Proprit du window */
    XSetStandardProperties (he_display, hn->win, "No name", "No name",
	0, he_argv, he_argc, NULL);
    HeWMGroup (hn);
    HeWMFocus (hn); 
    HeWMProtocols (hn);
    HeWMClass (hn); 
    HeEventMask (hn->win);
    HeWinListInsert (hn);

    /* Fin succes */
    return hn;
}


/*
 * Set
*/

void HeSetFrameLabel (He_node *hn, const char *nom)
{
    char bla[12];
    
    if (HeNotFrame(hn)) return;
    
    if (nom == NULL) nom = he_empty_str;
    strncpy (bla, nom, 12);
    bla[12-1] = 0;
    XSetStandardProperties (he_display, hn->win, nom, bla,
	0, he_argv, he_argc, NULL);
}

void HeSetFrameCloseProc (He_node *hn, He_close_proc proc)
{
    if (HeNotFrame(hn)) return;   
    hn->OBJ_UNION->close_proc = proc;
}

void HeSetFrameResizeProc (He_node *hn, He_resize_proc proc)
{
    if (HeNotFrame(hn)) return;
    hn->OBJ_UNION->resize_proc = proc;
}

void HeSetFrameReplaceProc (He_node *hn, He_replace_proc proc)
{
    if (HeNotFrame(hn)) return;
    hn->OBJ_UNION->replace_proc = proc;
}


/*
 * Get
*/

He_close_proc HeGetFrameCloseProc (He_node *hn)
{
    if (HeNotFrame(hn)) return NULL;
    return hn->OBJ_UNION->close_proc;
}
 
He_resize_proc HeGetFrameResizeProc (He_node *hn)
{
    if (HeNotFrame(hn)) return NULL;
    return hn->OBJ_UNION->resize_proc;
}
 
He_replace_proc HeGetFrameReplaceProc (He_node *hn)
{
    if (HeNotFrame(hn)) return NULL;
    return hn->OBJ_UNION->replace_proc;
}


/*
 * Applique la geometrie "wxh+-x+-y" sur le frame.
*/

void HeSetFrameGeom (He_node *hn, const char *geom)
{
    int m, x, y;
    unsigned int w, h;
    
    if (HeNotFrame(hn) || geom == NULL) return;
    
    m = XParseGeometry (geom, &x, &y, &w, &h);
    
    if (HE_BIT_ISSET (m, WidthValue) && w != 0 && w != hn->xm)
	HeSetWidth (hn, w);
        
    if (HE_BIT_ISSET (m, HeightValue) && h != 0 && h != hn->ym)
	HeSetHeight(hn, h);
    
    if (HE_BIT_ISSET (m, XNegative)) 
        HeSetX (hn, he_root_width + x 
            - HeGetWidth(hn) - he_wmdeco_width);
    else if (HE_BIT_ISSET (m, XValue)) 
        HeSetX (hn, x);
        
    if (HE_BIT_ISSET (m, YNegative)) 
        HeSetY (hn, he_root_height + y 
            - HeGetHeight(hn) - he_wmdeco_height);
    else if (HE_BIT_ISSET (m, YValue)) 
        HeSetY (hn, y);
}


/*
 * Met la propriete WM_TRANSIENT_FOR sur le frame hn.
 * Utile pour les dialogue et fenetre temporaires.
*/

void HeTransientFrame (He_node *hn)
{
    if (HeNotFrame(hn)) return;
    
    if (he_main_frame)
	XSetTransientForHint (he_display, hn->win, he_main_frame->win);
        
    else /* On ne connait pas encore he_main_frame ; on differe dans
            l'espoir que he_main_frame sera connu lors du update ..*/
	 HeUpdateMark (hn, HE_UPD_TRANSIENT);
}


/*
 *  Teste nature du noeud ; renvoie FALSE si erreur ; silencieux.
*/

int HeIsFrame (He_node *hn)
{
    return (hn == NULL || hn->meth != &he_meth_frame) ? FALSE : TRUE;
}


/*-------------------- M E T H O D E S - P R I V E E S ------------------------
 *
 * On ne teste pas si hn == NULL.
*/

/*
 * Initialisation du module ; appel par HeCreateFrame().
*/

void HeFrameAutoInit ()
{
    static int did_init = FALSE;
    
    if (did_init) return;
    
    HeMethodDefault (&he_meth_frame, HE_NAT_FRAME);
    he_meth_frame.props   = HE_PM_WIN_OWNER | HE_PM_TOP_WIN | HE_PM_STORE_FOCUS;
    he_meth_frame.destroy = HeFrameDestroy;
    he_meth_frame.fit     = HeFrameFit;
    he_meth_frame.update  = HeFrameUpdate;
    he_meth_frame.event   = HeFrameEvent;
    he_meth_frame.focus   = HeFrameStoreFocus;
    
    did_init = TRUE;
}


/*
 * Destruction d'un frame ; appele' par HeDestroyRec().
*/

void HeFrameDestroy (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    
    HeWinListRemove (hn);
    XDestroyWindow (he_display, hn->win);

    free (ob);
}


/*
 * Ajuste la taille du frame en fonction de son contenu.
*/

void HeFrameFit (He_node *hn)
{
    int xmax, ymax;

    if (HeNotFrame(hn)) return;
    
    HeBoundingBox (hn, &xmax, &ymax);
    HeSetWidth (hn, xmax);
    HeSetHeight(hn, ymax);
}


/*
 * M-A-J geometrie et affichage ; appele' par HeUpdateFlush().
 *
 * RQ - a la creation, il n'y a pas de HE_UPD_POS ; donc il n'y a
 * pas de Hint sur la position, ce qui laisse le WM placer la fenetre
 * ou il veut --> tres bien !!
 *    - des que le frame recoit une position (HeSetX ou ConfigureNotify)
 * alors on rajoute un tel Hint, et le WM respecte cette position lorsque 
 * la fenetre est remappee.
*/

void HeFrameUpdate (He_node *hn, He_upd_info *ui)
{
    HeFrameTopWinGeom (hn, ui);
        
    if (HE_BIT_ISSET (ui->mask, HE_UPD_POS))  HeWMHintPos (hn);
    if (HE_BIT_ISSET (ui->mask, HE_UPD_SIZE)) HeWMHintSize (hn);
    
    if (HE_BIT_ISSET (ui->mask, HE_UPD_TRANSIENT) && he_main_frame)
	XSetTransientForHint (he_display, hn->win, he_main_frame->win);

    if (HE_BIT_ISSET (ui->mask, HE_UPD_DRAW)) {
	if (HE_BIT_ISSET (hn->status, HE_STA_SHOW))
	     XMapRaised   (he_display, hn->win);
	else XUnmapWindow (he_display, hn->win);
    }
}


/*
 * Gestion evenements X ; appele' par HeEventDispatch().
 * Renvoie -1 si erreur.
*/

int HeFrameEvent (He_node *hn, He_event *hev)
{

#ifdef DEBUG
    if (hev->xev->type != MotionNotify)
        printf ("frame %s\n", HeGetXEventName (hev->type)); 
#endif

    switch (hev->type) {

	/* 
	** Touche clavier enfonce'e/relache'e : envoie hev au focus.
	*/
	case KeyPress :	
	case KeyRelease :	
            HeSendEventToFocus (hn, hev);
            break;
            
        /*
        ** Le WM informe que le win est reparente' ; on se sert de cet appel 
        **    - pour calculer la taille de la deco
        **    - pour mettre l'atome HELIUM_WMDECO a jour
        **    - pour completer l'option "-geom"
        **
        ** RQ : xreparent->x,y sont inexploitables.
        */    
	case ReparentNotify :	
            /* Appel unique */
            if (hn == he_main_frame && !he_wmdeco_test)
                HeFrameWMDeco (&hev->xev->xreparent);
            break;
            
	/*
	** Masquage d'un Window 
	*/
	case UnmapNotify : {

	    /* M-a-j des flags */
	    HE_BIT_CLR (hn->status, HE_STA_SHOW);
            HE_BIT_CLR (hn->status, HE_STA_MAP);
            
            /* Iconification gnrale : cache toutes les fenetres mappe'es.
	     * Les XUnmapWindow() vont generer des ev UnmapNotify, que
	     * l'on caracterise avec he_iconifying.
            */
	    if (hn == he_main_frame) {  

		int i;
		He_node *hop;

		he_iconifying = TRUE;

		for (i = 0; (hop = HeNodeScan(i)) != NULL; i++)
		if (HE_BIT_ISSET (hop->meth->props, HE_PM_TOP_WIN) && 
                    HE_BIT_ISSET (hop->status, HE_STA_SHOW))
                {
                    HE_BIT_SET (hop->status, HE_STA_ICONIC);
		    XUnmapWindow (he_display, hop->win);
                }
	    } 
	} break;


	/* 
	** Affichage d'un Window
	*/
	case MapNotify : {

	    /* M-a-j des flags.
             * RQ : HE_STA_SHOW est mis immdiatement par HeSetShow()
             *      HE_STA_MAP  est mis lors du mappage effectif.
            */
	    HE_BIT_SET (hn->status, HE_STA_SHOW);
            HE_BIT_SET (hn->status, HE_STA_MAP);
            
	    /* Reaffiche toutes les fenetres qui ont ete cachees 
	     * lors de l'iconification gnrale.
	    */
	    if (hn == he_main_frame && he_iconifying) {  
            
		int i;
		He_node *hop;

		he_iconifying = FALSE;
                
		for (i = 0; (hop = HeNodeScan(i)) != NULL; i++)
		if (HE_BIT_ISSET (hop->meth->props, HE_PM_TOP_WIN) && 
                    HE_BIT_ISSET (hop->status, HE_STA_ICONIC))
                {
                    HE_BIT_CLR (hop->status, HE_STA_ICONIC);
    		    XMapWindow (he_display, hop->win);
                }
	    }
            
            /* Dtection des REPLACE : cf commentaires dans ConfigureNotify */
            HeFrameDetectReplace (hn);
	} break;


	/* Intercepte la fermeture d'un Window 
	**
	** Genere appel a la CLOSE_PROC.
	*/
	case ClientMessage : {

	    OBJ_TYPE *ob = hn->OBJ_UNION;

	    /* On appelle la close_proc si il y en a une */
	    if (ob->close_proc != NULL)
		(ob->close_proc)(hn);
	    else if (hn == he_main_frame)
		 HeQuit (0);
	    else HeSetShow (hn, FALSE);
	} break;


	/* Changement de la geometrie d'un Window 
	**
	** On met a jour les parametres suite a un ev souris ou un HeSet()
	** On appelle la RESIZE_PROC puis la REPLACE_PROC.
	**
	** PB : selon le Window Magager (WM), il peut y avoir plusieurs 
	** evt ConfigureNotify a la suite, avec un border_width different
	** ou des coords x,y fantaisistes ; en effet les WM rajoutent 1
	** ou 2 parent-window autour du frame pour la deco, et les coords x,y 
	** sont souvent (mais pas toujours) relatives au parent quand le 
        ** frame est mappe', ou a root_win quand il n'est pas mappe'.
	** ON NE PEUT DONC PAS TENIR COMPTE DE x,y.
        **
        ** D'autre part on ne peut pas prendre en compte la position lorsque
        ** le frame n'est pas mapp. Or l'evenement MapNotify arrive souvent
        ** aprs les ConfigureNotify --> on fait aussi le test dans MapNotify.
	** 
	** NB dans un subwindow il n'y a pas ces problemes.
	*/
	case ConfigureNotify : {
                 
            /* Dtection des RESIZE */
	    HeFrameDetectResize (hn, 
                hev->xev->xconfigure.width, 
                hev->xev->xconfigure.height);
                
            /* Dtection des REPLACE */
            HeFrameDetectReplace (hn);
	} break;


	/* Changement de focus (RQ : a la cration du frame, 
         * on appelle HeWMFocus() pour que le WM envoie ces evt).
	*/
	case FocusIn  : HeFocusReceiveEvent (hn, TRUE);  break;
	case FocusOut : HeFocusReceiveEvent (hn, FALSE); break;
        
	/* Demande de refresh
	** On redessine le fond.
	*/
	case Expose : {
	/*
	    HeDrawBg (hn, he_col_bg);
	*/
	} break;

    } /* switch */

    return 0;
}


/*
 * Mthode renvoyant &focus_info.
*/

He_focus_info *HeFrameStoreFocus (He_node *hn)
{
    return &hn->OBJ_UNION->focus_info;
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Calcule les he_wmdeco_x,y lors du premier ReparentNotify,
 * puis met l'atome HELIUM_WMDECO a jour et complete l'option "-geom".
*/

void HeFrameWMDeco (XReparentEvent *ev)
{
    Window root, parent2, *childs;
    int x0, y0, x1, y1, x2, y2;
    unsigned int w0, h0, b0, w1, h1, b1, w2, h2, b2, depth, nchilds;
    
    /* Sauve tailles lues avec HeFrameReadAtomWMDeco() */
    int svg_xa = he_wmdeco_xa,    svg_ya = he_wmdeco_ya, 
        svg_xb = he_wmdeco_xb,    svg_yb = he_wmdeco_yb, 
        svg_w  = he_wmdeco_width, svg_h  = he_wmdeco_height; 
        
    /* Demande la geometrie du window et de son parent */
    if ( XGetGeometry (he_display, ev->window, &root, 
		&x0, &y0, &w0, &h0, &b0, &depth) == 0) return;
    if ( XGetGeometry (he_display, ev->parent, &root, 
		&x1, &y1, &w1, &h1, &b1, &depth) == 0) return;
   
    /* Certains WM rajoutent encore un deuxieme parent */
    parent2 = root;
    if (XQueryTree (he_display, ev->parent, &root, &parent2,
		&childs, &nchilds) != 0) {
	if (childs) XFree (childs);
    }
    
    if (parent2 == root) {  /* un seul parent */
    
        he_wmdeco_xa = x0 + b1;
        he_wmdeco_ya = y0 + b1;
        he_wmdeco_xb = b1;
        he_wmdeco_yb = b1;
        he_wmdeco_width  = w1 + 2*b1 - (w0 + 2*b0);
        he_wmdeco_height = h1 + 2*b1 - (h0 + 2*b0);
        he_wmdeco_test = TRUE;
        
    } else {  /* deux parents */
    
	if ( XGetGeometry (he_display, parent2, &root, 
		&x2, &y2, &w2, &h2, &b2, &depth) == 0) return;
                
        he_wmdeco_xa = x0 + b1 + x1 + b2;
        he_wmdeco_ya = y0 + b1 + y1 + b2;
        he_wmdeco_xb = he_wmdeco_xa;
        he_wmdeco_yb = he_wmdeco_ya;
        he_wmdeco_width  = w2 + 2*b2 - (w0 + 2*b0);
        he_wmdeco_height = h2 + 2*b2 - (h0 + 2*b0);
        he_wmdeco_test = TRUE;
    }
    
    /* Infos de dbugage */
    if (he_debug_option && strcmp(he_debug_option, "wmdeco") == 0) {
         printf ("HeFrameWMDeco:\n");
         printf ("  window   %4d ,%4d   %4d x%4d   %d\n", x0, y0, w0, h0, b0);
         printf ("  parent1  %4d ,%4d   %4d x%4d   %d\n", x1, y1, w1, h1, b1);
         if (parent2 != root)
         printf ("  parent2  %4d ,%4d   %4d x%4d   %d\n", x2, y2, w2, h2, b2);
         printf ("  WMDECO   %4d ,%4d   %4d x%4d   (%d,%d)\n\n",
             he_wmdeco_xa, he_wmdeco_ya, he_wmdeco_width, he_wmdeco_height,
             he_wmdeco_xb, he_wmdeco_yb);
    }

    /* Teste si l'atome HELIUM_WMDECO est a jour */
    if (svg_xa != he_wmdeco_xa    || svg_ya != he_wmdeco_ya || 
        svg_xb != he_wmdeco_xb    || svg_yb != he_wmdeco_yb || 
        svg_w  != he_wmdeco_width || svg_h  != he_wmdeco_height )
    {
        /* Complete l'option "-geom" */
        HeFrameOptionGeom ();
        
	/* Cree ou change l'atome */
        HeFrameWriteAtomWMDeco ();
    }
    
}


/*
 * Detection d'un resize (appel lors d'un ConfigureNotify)
*/

void HeFrameDetectResize (He_node *hn, int width, int height)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;

    if (ob->old_xm != width || ob->old_ym != height) {
    
        /* m-a-j */
	ob->old_xm = hn->xm = width;
	ob->old_ym = hn->ym = height; 
        
        HeWMHintSize (hn);
        
	if (ob->resize_proc != NULL)
	    (ob->resize_proc)(hn, hn->xm, hn->ym);
    }
}


/* Detection d'un replace (uniquement si le frame est mapp) :
 * on cherche les coord absolues du coin interieur ou exterieur du
 * frame / ecran. Si he_wmdeco_xa,y0 != 0 on obtient le coin
 * exterieur.
*/

void HeFrameDetectReplace (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    Window child;
    int nx, ny;
   
    if ( HE_BIT_ISSET (hn->status, HE_STA_MAP) &&
         XTranslateCoordinates (he_display, hn->win, he_root_window,
           -he_wmdeco_xa, -he_wmdeco_ya, &nx, &ny, &child) &&
         (ob->old_xb != nx || ob->old_yb != ny) )
    {
        /* m-a-j */
        ob->old_xb = hn->xb = nx;
        ob->old_yb = hn->yb = ny;
        
        /* On ne place le Hint de position que si les coords de la
           deco sont connues */
        if (he_wmdeco_test) HeWMHintPos (hn);

        if (ob->replace_proc != NULL)
	    (ob->replace_proc)(hn, hn->xb, hn->yb);
    }
}


/*
 * M-a-j geometrie du frame : on reporte les changements  son window.
 * Ceci remplace HeUpdateWinGeom() qui ne sert que pour les sub-windows.
 *
 * Appele' par HeFrameUpdate
*/

void HeFrameTopWinGeom (He_node *hn, He_upd_info *ui)
{
    unsigned int   mc = 0;  /* mask changes */
    XWindowChanges wc;

    if (hn == NULL || ui == NULL) return; 

    if (HE_BIT_ISSET (ui->mask, HE_UPD_DIMX)) { 
        HE_BIT_SET (mc, CWX); wc.x = hn->xb + he_wmdeco_xb;
    }
    if (HE_BIT_ISSET (ui->mask, HE_UPD_DIMY)) { 
        HE_BIT_SET (mc, CWY); wc.y = hn->yb + he_wmdeco_yb;
    }
    if (HE_BIT_ISSET (ui->mask, HE_UPD_DIMW)) { 
        HE_BIT_SET (mc, CWWidth);  wc.width  = hn->xm;
    }
    if (HE_BIT_ISSET (ui->mask, HE_UPD_DIMH)) { 
        HE_BIT_SET (mc, CWHeight); wc.height = hn->ym;
    }

    if (mc != 0) 
	/* C'est un top-level window, reparente' par le WM */
	XReconfigureWMWindow (he_display, hn->win, he_screen, mc, &wc);
}


/*
 * Applique l'option "-geom" sur he_frame_principal.
 *
 * 2 appels :
 *   - avant que la fenetre ne soit mappee.
 *   - une fois que la taille de la deco est connue
*/

void HeFrameOptionGeom ()
{
    /* On memorise le parametre de l'option -geom au premier appel */
    if (!he_option_geom) he_option_geom = HeGetOption ("-geom", NULL);
    
    /* Il n'y a pas d'option -geom */
    if (!he_option_geom) return;
    
    HeSetFrameGeom (he_main_frame, he_option_geom);
}


/*
 * Lit l'atome HELIUM_WMDECO dans he_root_window
 *
 * On appelle cette fonction 2 fois :
 * - une premire fois dans HeInit() : de la sorte on peut se servir tout
 *   de suite de he_wmdeco_width et he_wmdeco_height.
 * - une seconde fois dans HeFrameWMDeco pour ventuellement crer/corriger
 *   l'atome.
*/

void HeFrameReadAtomWMDeco ()
{
    Atom type = 0;
    Ulong nitems = 0;
    Uchar *data = NULL;
    int format = 0, err;
    
    err = HeAtomGetProp (he_root_window, he_atom_wmdeco, False,
	&type, &format, &nitems, &data);
        
    if (err == 0 && type == XA_STRING && format == 8) {
    
#ifdef DEBUG
    printf ("Read HELIUM_WMDECO = \"%s\"\n", data);
#endif
        sscanf ((char *) data, "%d %d %d %d %d %d",
	    &he_wmdeco_xa, &he_wmdeco_ya, &he_wmdeco_xb, &he_wmdeco_yb, 
            &he_wmdeco_width, &he_wmdeco_height);
    }
    
    if (data) XFree (data);
}


/*
 * Cree/ecrit l'atome HELIUM_WMDECO dans he_root_window
*/

void HeFrameWriteAtomWMDeco ()
{
    char bla[80];
    
    sprintf (bla, "%d %d %d %d %d %d",
	he_wmdeco_xa, he_wmdeco_ya, he_wmdeco_xb, he_wmdeco_yb, 
        he_wmdeco_width, he_wmdeco_height);
        
#ifdef DEBUG
    printf ("Write HELIUM_WMDECO = \"%s\"\n", bla);
#endif

    XChangeProperty (he_display, he_root_window, he_atom_wmdeco,
	XA_STRING, 8, PropModeReplace, (Uchar *) bla, strlen(bla));
}


/*
 *  Teste nature du noeud ; renvoie TRUE si erreur. 
*/

int HeNotFrame (He_node *hn)
{
    if (hn == NULL || hn->meth != &he_meth_frame) {
	HeError ("Frame expected\n");
        return TRUE;
    }
    return FALSE;
}

