/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/geobase.c - 31/05/1996
 *
 * 
*/

#include <helium.h>


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Set
 *
 * ATTENTION, toujours placer HeUpdateMark() AVANT un changement de coords.
*/

void HeSetShow (He_node *hn, int show)
{
    if (hn == NULL) return; 
    HeUpdateMark (hn, HE_UPD_DRAW);
    if (show)
         HE_BIT_SET (hn->status, HE_STA_SHOW);  
    else HE_BIT_CLR (hn->status, HE_STA_SHOW);  
}

void HeSetActive (He_node *hn, int active)
{
    if (hn == NULL) return; 
    HeUpdateMark (hn, HE_UPD_DRAW);
    if (active)
	 HE_BIT_SET (hn->status, HE_STA_ACTIVE);
    else HE_BIT_CLR (hn->status, HE_STA_ACTIVE);
}

void HeSetX (He_node *hn, int x)
{
    if (hn == NULL || hn->xb == x) return; 
    HeUpdateMark (hn, HE_UPD_DIMX);
    hn->xb = x;  
}

void HeSetY (He_node *hn, int y)
{
    if (hn == NULL || hn->yb == y) return; 
    HeUpdateMark (hn, HE_UPD_DIMY);
    hn->yb = y;  
}

void HeSetWidth (He_node *hn, int width)
{
    if (hn == NULL || width <= 0 || width >= 65536 || hn->xm == width) return; 
    HeUpdateMark (hn, HE_UPD_DIMW);
    hn->xm = width;  
}

void HeSetHeight (He_node *hn, int height)
{
    if (hn == NULL || height <= 0 || height >= 65536 || hn->ym == height) return; 
    HeUpdateMark (hn, HE_UPD_DIMH);
    hn->ym = height;  
}

void HeSetBorder (He_node *hn, int border)
{
    if (hn == NULL) return; 
    HeUpdateMark (hn, HE_UPD_BORDER);
    hn->win_border = border;  
}

void HeSetClientData (He_node *hn, void *client_data)
{
    if (hn == NULL) return; 
    hn->client_data = client_data;  
}

void HeSetDestroyProc (He_node *hn, He_destroy_proc proc)
{
    if (hn == NULL) return; 
    hn->destroy_proc = proc;  
}


/*
 * Get
*/

He_node *HeGetOwner (He_node *hn)
{
    if (hn == NULL) return NULL; 
    return hn->owner;  
}

He_node *HeGetNext (He_node *hn)
{
    if (hn == NULL) return NULL; 
    return hn->next_node;  
}

He_node *HeGetSub (He_node *hn)
{
    if (hn == NULL) return NULL; 
    return hn->sub_node;  
}

int HeGetShow (He_node *hn)
{
    if (hn == NULL) return FALSE; 
    return HE_BIT_ISSET (hn->status, HE_STA_SHOW);  
}

int HeGetActive (He_node *hn)
{
    if (hn == NULL) return FALSE; 
    return HE_BIT_ISSET (hn->status, HE_STA_ACTIVE);
}

int HeGetX (He_node *hn)
{
    if (hn == NULL) return 0; 
    return hn->xb;  
}

int HeGetY (He_node *hn)
{
    if (hn == NULL) return 0; 
    return hn->yb;  
}

int HeGetWidth (He_node *hn)
{
    if (hn == NULL) return 0; 
    return hn->xm;  
}

int HeGetHeight (He_node *hn)
{
    if (hn == NULL) return 0; 
    return hn->ym;  
}

int HeGetBorder (He_node *hn)
{
    if (hn == NULL) return 0; 
    return hn->win_border;  
}

void *HeGetClientData (He_node *hn)
{
    if (hn == NULL) return NULL; 
    return hn->client_data;  
}

He_destroy_proc HeGetDestroyProc (He_node *hn)
{
    if (hn == NULL) return NULL; 
    return hn->destroy_proc;  
}

Window HeGetWindow (He_node *hn)
{
    if (hn == NULL) return 0; 
    return hn->win;  
}

Ulong HeGetNat (He_node *hn)
{
    if (hn == NULL || hn->meth == NULL) return 0L; 
    return hn->meth->nat;  
}

int HeGetId (He_node *hn)
{
    if (hn == NULL) return 0; 
    return hn->id;  
}


/*
 * Renvoie la largeur ou hauteur totale de hn
*/

int HeGetExtWidth (He_node *hn)
{
    if (hn == NULL) return 0; 
    if (HeIsFrame(hn)) return HeGetWidth(hn) + he_wmdeco_width;
    return HeGetWidth(hn) + 2*HeGetBorder(hn);
}

int HeGetExtHeight (He_node *hn)
{
    if (hn == NULL) return 0; 
    if (HeIsFrame(hn)) return HeGetHeight(hn) + he_wmdeco_height;
    return HeGetHeight(hn) + 2*HeGetBorder(hn);
}


/*
 * Renvoie les coords du coin en bas  droite, extrieur  hn
*/

int HeGetExtX (He_node *hn)
{
    return HeGetX(hn) + HeGetExtWidth(hn);
}

int HeGetExtY (He_node *hn)
{
    return HeGetY(hn) + HeGetExtHeight(hn);
}


/*
 * Incrmente X, Y, Width ou Height
*/

void HeMoveX (He_node *hn, int dx)
{
    HeSetX (hn, HeGetX (hn) + dx);
}

void HeMoveY (He_node *hn, int dy)
{
    HeSetY (hn, HeGetY (hn) + dy);
}

void HeMoveWidth (He_node *hn, int dwidth)
{
    HeSetWidth (hn, HeGetWidth (hn) + dwidth);
}

void HeMoveHeight (He_node *hn, int dheight)
{
    HeSetHeight (hn, HeGetHeight (hn) + dheight);
}


/*
 * Justification par rapport  ref, ou au bord du pre si ref est NULL.
 * L'espacement intra-panel, extra-panel ou extra-frame est automatique.
 *
 * RQ : si ref et hn n'ont pas le mme pre, les fonctions sortent avec
 *      un message d'erreur pour viter un rsultat alatoire : en effet,
 *      les coord sont relatives au pre.
*/

void HeJustify (He_node *hn, He_node *ref, int pos)
{
    int x, y, bx, by;
    
    if (hn == NULL) return; 
    
    /* Test si ref et hn n'ont pas le mme pre */ 
    if (ref != NULL && hn->owner != ref->owner) {
	HeError ("HeJustify(hn,ref,pos):"
                 " hn and ref do not have the same owner\n");
        return;
    }
    
    /* Calcul de l'espacement  */
    bx = (HeIsPanel(hn->owner)) ? (ref == NULL) ? 
	HeGetPanelXAlign(hn->owner) : HeGetPanelXGap(hn->owner) : 0;
    by = (HeIsPanel(hn->owner)) ? (ref == NULL) ? 
	HeGetPanelYAlign(hn->owner) : HeGetPanelYGap(hn->owner) : 0;
    
    /* Calcul de la position horizontale */
    switch (pos) {
        case HE_LEFT : 
        case HE_TOP_LEFT : 
        case HE_BOTTOM_LEFT : 
	    x = (ref == NULL) ? 0 : HeGetExtX(ref);
	    HeSetX (hn, x+bx);
            break;
        case HE_RIGHT : 
        case HE_TOP_RIGHT : 
        case HE_BOTTOM_RIGHT : 
	    x = (ref == NULL) ? HeGetWidth(hn->owner) : HeGetX(ref);
	    HeSetX (hn, x - bx - HeGetExtWidth(hn));
            break;
    }
    
    /* Calcul de la position verticale */
    switch (pos) {
        case HE_TOP : 
        case HE_TOP_LEFT : 
        case HE_TOP_RIGHT : 
	    y = (ref == NULL) ? 0 : HeGetExtY(ref);
	    HeSetY (hn, y+by);
            break;
        case HE_BOTTOM : 
        case HE_BOTTOM_LEFT :
        case HE_BOTTOM_RIGHT :
	    y = (ref == NULL) ? HeGetHeight(hn->owner) : HeGetY(ref);
	    HeSetY (hn, y - by - HeGetExtHeight(hn));
            break;
    }
}


/*
 * Elargissement jusqu' ref, ou jusqu'au bord du pre si ref est NULL.
 * L'espacement intra-panel, extra-panel ou extra-frame est automatique.
 *
 * RQ : si ref et hn n'ont pas le mme pre, les fonctions sortent avec
 *      un message d'erreur pour viter un rsultat alatoire : en effet,
 *      les coord sont relatives au pre.
*/

void HeExpand (He_node *hn, He_node *ref, int pos)
{
    int x, y, bx, by;
    
    if (hn == NULL) return; 
    
    /* Test si ref et hn n'ont pas le mme pre */ 
    if (ref != NULL && hn->owner != ref->owner) {
	HeError ("HeJustify(hn,ref,pos):"
	         " hn and ref do not have the same owner\n");
        return;
    }
    
    /* Calcul de l'espacement  */
    bx = (HeIsPanel(hn->owner)) ? (ref == NULL) ? 
	HeGetPanelXAlign(hn->owner) : HeGetPanelXGap(hn->owner) : 0;
    by = (HeIsPanel(hn->owner)) ? (ref == NULL) ? 
	HeGetPanelYAlign(hn->owner) : HeGetPanelYGap(hn->owner) : 0;
    
    /* Calcul de la position et de la taille horizontale */
    switch (pos) {
        case HE_LEFT : 
        case HE_TOP_LEFT : 
        case HE_BOTTOM_LEFT : 
	    x = (ref == NULL) ? 0 : HeGetExtX(ref); x += bx;
            HeMoveWidth (hn, HeGetX(hn) - x); HeSetX (hn, x);
            break;
        case HE_RIGHT : 
        case HE_TOP_RIGHT : 
        case HE_BOTTOM_RIGHT : 
	    x = (ref == NULL) ? HeGetWidth(hn->owner) : HeGetX(ref); 
            x -= bx; HeMoveWidth (hn, x - HeGetExtX(hn));
            break;
    }
    
    /* Calcul de la position et de la taille verticale */
    switch (pos) {
        case HE_TOP : 
        case HE_TOP_LEFT : 
        case HE_TOP_RIGHT : 
	    y = (ref == NULL) ? 0 : HeGetExtY(ref); y += by;
            HeMoveHeight (hn, HeGetY(hn) - y); HeSetY (hn, y);
            break;
        case HE_BOTTOM : 
        case HE_BOTTOM_LEFT :
        case HE_BOTTOM_RIGHT :
	    y = (ref == NULL) ? HeGetHeight(hn->owner) : HeGetY(ref); 
            y -= by; HeMoveHeight (hn, y - HeGetExtY(hn));
            break;
    }
}


/*
 * Calcule la boite englobante des sous-noeuds (un seul niveau).
 * Tient compte du status destroy.
 * Utilise dans les mthodes Fit.
*/

void HeBoundingBox (He_node *hn, int *xmax, int *ymax)
{
    He_node *hop;
    int x, y, w;
    
    /* Taille min 1x1 */
    *xmax = 1; *ymax = 1;
    
    if (hn == NULL) return;
    
    for (hop = hn->sub_node; hop != NULL; hop = hop->next_node) {

	/* On ignore hop s'il est a detruire */
        if (HE_BIT_ISSET (hop->status, HE_STA_DESTROY)) continue;

	/* on cherche le win_border */
	w = HeGetBorder (hop);

	/* On cherche le max en ajoutant w*2 */
	x = hop->xb + hop->xm + w*2; if (*xmax < x) *xmax = x;
	y = hop->yb + hop->ym + w*2; if (*ymax < y) *ymax = y;
    }
}


/*
 * Ajuste la taille du widget  son contenu
*/

void HeFit (He_node *hn)
{
    if (hn == NULL) return; 
    (hn->meth->fit)(hn);    
}

void HeSetAutoFit (He_node *hn, int autofit)
{
    if (hn == NULL) return; 
    if (autofit)
         HE_BIT_SET (hn->status, HE_STA_AUTOFIT);  
    else HE_BIT_CLR (hn->status, HE_STA_AUTOFIT);  
}

int HeGetAutoFit (He_node *hn)
{
    if (hn == NULL) return FALSE; 
    return HE_BIT_ISSET (hn->status, HE_STA_AUTOFIT);  
}


/*
 * Translate les coordonnes src_x, src_y par rapport au coin intrieur en haut
 *  gauche de ref en coordonnes absolues abs_x, abs_y par rapport  l'cran.
 * ref peut tre n'importe quel widget ou gadget.
 * Si ref est NULL ou he_root_node, les coordonnes restent inchanges.
*/

void HeAbsoluteCoords (He_node *ref, int src_x, int src_y, 
	int *abs_x, int *abs_y)
{
    if (ref == NULL || ref == he_root_node) {
	*abs_x = src_x; *abs_y = src_y;
    } else {
	Window child;
        
	if (!HE_BIT_ISSET (ref->meth->props, HE_PM_WIN_OWNER)) {
	    src_x += ref->xb + ref->win_border; 
            src_y += ref->yb + ref->win_border;
	}
         
	XTranslateCoordinates (he_display, ref->win, he_root_window,
		src_x, src_y, abs_x, abs_y, &child);
    } 
}


/*
 * Calcule les coordonnes actuelles de la souris relativement  hn,
 * ou absolues par rapport  l'cran si hn est nul.
 * Renvoie TRUE en cas de succs, FALSE sinon (pas le mme screen, etc). 
*/

int HeQueryPointerCoords (He_node *hn, int *x, int *y)
{
    Window root, child;
    int rootx, rooty;
    unsigned int mask;
    
    if (XQueryPointer (he_display, 
    		hn == NULL ? he_root_window : hn->win, 
                &root, &child, &rootx, &rooty,
    		x, y, &mask) == False)
        return FALSE;
        
    if (hn != NULL && !HE_BIT_ISSET (hn->meth->props, HE_PM_WIN_OWNER)) {
        /* On suppose que hn->owner est WIN_OWNER */
        *x -= hn->xb + hn->win_border;
        *y -= hn->yb + hn->win_border;
    }
    
    return TRUE;
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialisation des champs d'un He_node.
 * Appele' par HeNodeNew().
*/

void HeBaseInit (He_node *hn, int id)
{
    HeBaseZero (hn);

    hn->owner = hn->next_node = hn->sub_node = (He_node *) NULL;
    hn->id = id;
    hn->xb = hn->yb = 0;
    hn->xm = hn->ym = 50;
    
    HE_BIT_SET (hn->status, HE_STA_SHOW);
    HE_BIT_SET (hn->status, HE_STA_ACTIVE);
    HE_BIT_SET (hn->status, HE_STA_AUTOFIT);
    
    hn->upd_index = -1;  /* Rien  updater */
    hn->meth = &he_meth_void;
    hn->obj.any = (void *) NULL;
    hn->client_data = (void *) NULL;
    hn->destroy_proc = NULL;

    hn->win = 0;
    hn->win_border = 0;
}


/*
 * RAZ des champs d'un node.
*/

void HeBaseZero (He_node *hn)
{
    memset (hn, 0, sizeof (He_node));
}


/*
 * Creation d'un widget.
 * Renvoie ptr succes, ou NULL erreur.
*/

He_node *HeBaseCreate (He_node *owner, He_method *meth)
{
    He_node *ho, *hn;

    /* Teste si meth est ok */
    if (meth->magic != HE_METH_MAGIC) {
	HeError ("HeBaseCreate: meth has a bad magic number.\n");
	return NULL;
    } 
    
    /* Retrouve owner, ou le remplace par he_root_node */
    ho = (owner == NULL) ? he_root_node : owner;

    /* Cree le noeud hn, le rattache a ho */
    hn = HeNodeNew ();
    if (hn == NULL) return NULL;
    HeNodeAddTree (ho, hn);

    /* Fixe hn->meth */
    hn->meth = meth;

    return hn;
}


/*
 * Teste si (x,y) est dans le rectangle de hn
 * Renvoie TRUE ou FALSE 
 * Si hn est NULL, renvoie FALSE.
*/

int HeIsClic (He_node *hn, int x, int y)
{
    if (hn == NULL) return FALSE;
    return ( HE_BIT_ISSET (hn->status, HE_STA_SHOW) && 
             hn->xb <= x && hn->yb <= y && 
	     x < hn->xb+hn->xm && y < hn->yb+hn->ym ) 
	? TRUE : FALSE;
}


/*
 * Cherche le hn pour lequel (x,y) est dans son rectangle.
 * Selon propagate, ne cherche que dans les next_node (FALSE),
 * ou cherche aussi dans les sub_node (TRUE).
 * Ne cherche pas dans les noeuds qui sont WIN_OWNER, car ils
 * reoivent l'evt directement.
*/

He_node *HeFindClic (He_node *prem, int x, int y, int propagate)
{
    He_node *hn;
    
    /* Parcourt les frres */
    for (hn = prem; hn != NULL; hn = hn->next_node)
        if ( HE_BIT_ISSET (hn->status, HE_STA_SHOW) &&
             !HE_BIT_ISSET (hn->meth->props, HE_PM_WIN_OWNER) )
        {
            /* Premier niveau */
            if ( hn->xb <= x && hn->yb <= y &&
                 x < hn->xb+hn->xm && y < hn->yb+hn->ym )
            return hn;
            
	    /* Cherche rcursivement dans les sous-niveaux */
            if (propagate == TRUE && hn->sub_node != NULL) {
                He_node *tmp = HeFindClic (hn->sub_node, x, y, propagate);
                if (tmp != NULL) return tmp;
            }
        }

    /* Rien trouv */
    return NULL;
}



