/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/menu.c - 11/01/2002
 *
 * 
*/

#include <helium.h>

/* Type et union du package 
*/
#define  OBJ_TYPE	Obj_menu
#define  OBJ_UNION	obj.menu

static He_method he_meth_menu;

/* Permet de savoir si au moins un menu est visible */
static int he_menu_visibility = FALSE;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 *  Cration d'un menu.
 *  Renvoie le noeud ou NULL.
*/

He_node *HeCreateMenu ()
{
    He_node *hn;
    OBJ_TYPE *ob;

    HeMenuAutoInit();

    /* On cree le noeud */
    hn = HeBaseCreate (NULL, &he_meth_menu);
    if (hn == NULL) return NULL;

    /* On cree l'objet */
    ob = hn->OBJ_UNION = malloc (sizeof (OBJ_TYPE));
    if (ob == NULL) {
	HeError ("HeCreateMenu: malloc error\n");
	return NULL;
    }

    /* Init des champs du noeud */
    HE_BIT_CLR (hn->status, HE_STA_SHOW);
    hn->win_border = 0;
    
    /* init des champs de l'objet */
    ob->label = NULL;
    ob->pointed = 0;
    ob->kbd_proc = NULL;
    
    HeFocusInitInfo (&ob->focus_info);

    /* Creation du Window avec Xlib */
    hn->win = XCreateSimpleWindow (he_display, 
	he_root_window,
	hn->xb, hn->yb, hn->xm, hn->ym, 
	hn->win_border,
	he_black,
	he_col_bg);

    /* Proprit du window */
    XSetStandardProperties (he_display, hn->win, "No name", "No name",
	0, he_argv, he_argc, NULL);
        
    /* HeWMGroup (hn);  ignor par le WM */
    /* HeWMFocus (hn);  ignor par le WM */
    
    HeWMProtocols (hn);
    HeWMClass (hn); 
    HeWMOverrideRedirect (hn);
    HeWMSaveUnder (hn);
    HeEventMask (hn->win);
    HeWinListInsert (hn);

    /* Update diffr */
    HeUpdateMark (hn, HE_UPD_DRAW);

    /* Fin succes */
    return hn;
}


/*
 * Set
*/

void HeSetMenuLabel (He_node *hn, const char *label)
{
    OBJ_TYPE *ob;
    
    if (HeNotMenu(hn)) return;
    ob = hn->OBJ_UNION;
    
    free (ob->label);
    ob->label = (label == NULL) ? NULL : strdup(label);
    
    if (label == NULL) {
	XSetStandardProperties (he_display, hn->win, "No name", "No name",
	    0, he_argv, he_argc, NULL);
    } else {
	char bla[12];
	strncpy (bla, label, 12);
	bla[12-1] = 0;
 	XSetStandardProperties (he_display, hn->win, label, bla,
	    0, he_argv, he_argc, NULL);
    }
        
    /* HeMenuFit (hn); */
}


/*
 * Get
*/

char *HeGetMenuLabel (He_node *hn)
{
    if (HeNotMenu(hn)) return he_empty_str;
    return hn->OBJ_UNION->label ? hn->OBJ_UNION->label : he_empty_str;
}


/*
 * Affiche le menu hn aux coordonnes x,y par rapport  ref,
 * ou par rapport  l'cran si ref est NULL ;
 * recadre le menu pour qu'il ne dpasse pas l'cran.
*/

void HeShowMenuXY (He_node *hn, He_node *ref, int x, int y)
{
    if (HeNotMenu(hn)) return;
    
    /* Le pre du Window du menu est le root Window ; il faut donc 
       exprimer les coordonnes x,y par rapport au root Window. */
    HeAbsoluteCoords (ref, x, y, &x, &y);

    /* On recadre le menu dans l'cran */
    if (x < 0) x = 0;
    else if (x + hn->xm > he_root_width) x = he_root_width - hn->xm;
    if (y < 0) y = 0;
    else if (y + hn->ym > he_root_height) y = he_root_height - hn->ym;

    /* Affichage en x,y */
    HeSetX (hn, x); HeSetY (hn, y); HeSetShow (hn, TRUE);
}


/*
 * Affiche le menu hn en dessous ou  droite de ref, selon layout ;
 * recadre le menu pour qu'il ne dpasse pas l'cran.
 *
 * layout = HE_BOTTOM : menu dessous ref (ou dessus).
 *          HE_RIGHT  : menu  droite (ou gauche) de ref, vers le bas.
*/

void HeShowMenu (He_node *hn, He_node *ref, int layout)
{
    int x, y, l = ref ? ref->xm : 0, h = ref ? ref->ym : 0;
    
    if (HeNotMenu(hn)) return;
    
    /* Coordonnes absolues */
    HeAbsoluteCoords (ref, 0, 0, &x, &y);

    switch (layout)
    {
        case HE_BOTTOM :
            /* On met en dessous, sinon au-dessus */
            y += h;
            if (y + hn->ym > he_root_height && y - h - hn->ym >= 0) 
                y = y - h - hn->ym;
            break;
            
        case HE_RIGHT :
            /* On met  droite, sinon  gauche */
            x += l;
            if (x + hn->xm > he_root_width && x - l - hn->xm >= 0)
                x = x - l - hn->xm;
            break;
            
        default :
            HeError ("HeShowMenu: bad layout\n");
	    return;
    }
    
    /* On recadre */
    if (x < 0) x = 0;
    else if (x + hn->xm > he_root_width) x = he_root_width - hn->xm;
    if (y < 0) y = 0;
    else if (y + hn->ym > he_root_height) y = he_root_height - hn->ym;

    /* Affichage en x,y */
    HeSetX (hn, x); HeSetY (hn, y); HeSetShow (hn, TRUE);
}


/*
 * Ferme menu_start et tous ses sous-menus qui sont ouverts.
*/

void HeHideMenus (He_node *menu_start)
{
    He_node *item;
    
    if (!HeIsMenu(menu_start)) return;
    
    if (!HE_BIT_ISSET (menu_start->status, HE_STA_SHOW)) return;
    HeSetShow (menu_start, FALSE);
    
    /* Rcursif */
    for (item = menu_start->sub_node; item != NULL; item = item->next_node)
        HeHideMenus ( HeGetMenuItemSubMenu (item) );
}


/*
 *  Teste nature du noeud ; renvoie FALSE si erreur ; silencieux.
*/

int HeIsMenu (He_node *hn)
{
    return (hn == NULL || hn->meth != &he_meth_menu) ? FALSE : TRUE;
}


/*-------------------- M E T H O D E S - P R I V E E S ------------------------
 *
 * On ne teste pas si hn == NULL.
*/

/*
 * Initialisation du module ; appel par HeCreateMenu().
*/

void HeMenuAutoInit ()
{
    static int did_init = FALSE;
    
    if (did_init) return;
    
    HeMethodDefault (&he_meth_menu, HE_NAT_MENU);
    he_meth_menu.props   = HE_PM_WIN_OWNER | HE_PM_TOP_WIN | HE_PM_STORE_FOCUS;
    he_meth_menu.destroy = HeMenuDestroy;
    he_meth_menu.fit     = HeMenuFit;
    he_meth_menu.update  = HeMenuUpdate;
    he_meth_menu.event   = HeMenuEvent;
    he_meth_menu.draw    = HeMenuDraw;
    he_meth_menu.focus   = HeMenuStoreFocus;
    
    did_init = TRUE;
}


/*
 * Destruction d'un menu ; appele' par HeDestroyRec().
*/

void HeMenuDestroy (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;

    HeWinListRemove (hn);
    XDestroyWindow (he_display, hn->win);

    free (ob->label);
    free (ob);
}


/*
 * Ajuste la taille du menu en fonction de son contenu.
*/

void HeMenuFit (He_node *hn)
{
    int xmax, ymax;
    if (HeNotMenu(hn)) return;
    
    HeBoundingBox (hn, &xmax, &ymax);
    HeSetWidth (hn, xmax + 1);
    HeSetHeight(hn, ymax + 1);
}


/*
 * M-A-J geometrie et affichage ; appele' par HeUpdateFlush().
*/

void HeMenuUpdate (He_node *hn, He_upd_info *ui)
{
    HeUpdateWinGeom (hn, ui);
        
    if (HE_BIT_ISSET (ui->mask, HE_UPD_BORDER)) {
	XSetWindowBorderWidth (he_display, hn->win, hn->win_border);
    }

    if (HE_BIT_ISSET (ui->mask, HE_UPD_DRAW)) {
	if (HE_BIT_ISSET (hn->status, HE_STA_SHOW)) {
	    XMapRaised   (he_display, hn->win);
            he_menu_visibility = TRUE;
	} else {
            XUnmapWindow (he_display, hn->win);
        }
    }
}


/*
 * Methode pour gerer les evenements X.
 * Renvoie -1 si erreur.
 *
 * Appele' par HeEventDispatch()
*/

int HeMenuEvent (He_node *hn, He_event *hev)
{
    OBJ_TYPE *ob = hn->OBJ_UNION ;

    if (hev->xev->type != MotionNotify)
        printf ("menu %s\n", HeGetXEventName (hev->xev->type));

    switch (hev->type) {
    
        case EnterNotify : 
            if (!HE_BIT_ISSET (hn->status, HE_STA_FOCUS_IN))
                XSetInputFocus (he_display, hn->win, RevertToNone, CurrentTime);
            break;

	case FocusIn  : HeFocusReceiveEvent (hn, TRUE);  break;
	case FocusOut : HeFocusReceiveEvent (hn, FALSE); break;
        
        case ButtonRelease : {  /* Temporaire pour tests */
            /*int res = XUngrabPointer (he_display, CurrentTime);
            printf ("XUngrabPointer : %d\n", res); */
        } break;
            
        /* Pour les survols de menu avec bouton enfonc, 
           modifier HeEventToGadgets  */
        
        case MapNotify : { 
            /* Rq : on ne peut faire le grab que lorsque le Window est mapp,
               c'est pourquoi on le fait seulement ici. */
	    /* int res = XGrabPointer (he_display, hn->win, False,
                ButtonPressMask | ButtonReleaseMask |
                EnterWindowMask | LeaveWindowMask | PointerMotionMask,
                GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
            printf ("XGrabPointer : %d\n", res);   */
        
            /* mettre aussi un timeout pour enlever le grab (show  false) */
        } break;

    } /* switch */

    return HeEventToGadgets (hn, hev, ob->kbd_proc, &ob->pointed);
}


/*
 * Dessin du menu.
 *
*/

void HeMenuDraw (He_node *hn)
{
    HeDrawRect3D (hn->win, 0, 0, hn->xm-1, hn->ym-1,
	he_col_light_bd, he_col_dark_bd);
}


/*
 * Mthode renvoyant &focus_info.
*/

He_focus_info *HeMenuStoreFocus (He_node *hn)
{
    return &hn->OBJ_UNION->focus_info;
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Filtre pour fermer tous les menus visibles en cas de clic  l'extrieur ;
 * Ceci evite de faire un XGrabPointer.
 * Appel par HeEventDispatch.
 * Renvoie TRUE (vnement filtr) ou FALSE (ignor).
*/

int HeMenuFilterEvents (He_node *hn, He_event *hev)
{
    if (hev->type != ButtonPress) return FALSE;
    
    if (HeIsMenu(hn) || HeIsMenuItem(hn)) return FALSE;
    
    if (!he_menu_visibility) return FALSE;
    
    if (HeMenuHideAll() == 0) return FALSE;
    
    printf ("HeMenuFilterEvents: ButtonPress Event filtered\n"); 
    return TRUE;
}


/*
 * Ferme tous les menus et rinitialise he_menu_visibility.
 * Renvoie le nombre de menus ferms.
 * Appel par HeMenuEventsFilter
*/

int HeMenuHideAll ()
{
    int i, count_closed = 0;
    He_node *hop;
    
    for (i = 0; (hop = HeNodeScan(i)) != NULL; i++) {
        if (HeIsMenu(hop) == TRUE && HE_BIT_ISSET (hop->status, HE_STA_SHOW)) {
            HeSetShow (hop, FALSE);
            count_closed++;
        }
    }
    
    printf ("HeMenuHideAll: %d menus closed\n", count_closed);
    
    he_menu_visibility = FALSE;
    return count_closed;
}


/*
 *  Teste nature du noeud ; renvoie TRUE si erreur. 
*/

int HeNotMenu (He_node *hn)
{
    if (hn == NULL || hn->meth != &he_meth_menu) {
	HeError ("Menu expected\n");
        return TRUE;
    }
    return FALSE;
}


