/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/menu_item.c - 24/01/2002
 *
 * 
*/

#include <helium.h>

/* Type et union du package 
*/
#define  OBJ_TYPE	Obj_menu_item
#define  OBJ_UNION	obj.menu_item

static He_method he_meth_menu_item;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 *  Cration d'un menu_item.
 *  Renvoie le noeud ou NULL.
*/

He_node *HeCreateMenuItem (He_node *owner)
{
    He_node *hn;
    OBJ_TYPE *ob;

    HeMenuItemAutoInit();

    /* On cree le noeud */
    hn = HeBaseCreate (owner, &he_meth_menu_item);
    if (hn == NULL) return NULL;

    /* On cree l'objet */
    ob = hn->OBJ_UNION = malloc (sizeof (OBJ_TYPE));
    if (ob == NULL) {
	HeError ("HeCreateMenuItem: malloc error\n");
	return NULL;
    }

    /* Init des champs du noeud */    
    hn->xm = hn->ym = 0;		/* Inhibe le 1er HeUpdateDrawOldBg */
    hn->win = hn->owner->win;

    /* init des champs de l'objet */
    ob->label = NULL;
    ob->rank = -1;       /* Plac  la fin */
    ob->submenu_id = 0;
    ob->notify_proc = NULL;

    /* Positionne l'objet en fonction de son rank */
    /* HeMenuPosMenuItem (hn);  ECRIRE !!! */

    /* Fixe taille par defaut, provoque HE_UPD_ANYDIM | HE_UPD_DRAW */
    HeMenuItemFit (hn);

    /* Fin succes */
    return hn;
}


/*
 * Creation d'un bouton parametre'
*/

He_node *HeCreateMenuItemP (He_node *owner, const char *label, 
                          He_notify_proc proc, void *client_data)
{
    He_node *tmp = HeCreateMenuItem (owner);
    HeSetMenuItemLabel (tmp, label);
    HeSetMenuItemNotifyProc (tmp, proc);
    HeSetClientData (tmp, client_data);
    return tmp;
}


/*
 * Set
*/

void HeSetMenuItemLabel (He_node *hn, const char *label)
{
    OBJ_TYPE *ob;
    
    if (HeNotMenuItem(hn)) return;
    ob = hn->OBJ_UNION;
    
    free (ob->label);
    ob->label = (label == NULL) ? NULL : strdup(label);
    
    if (HE_BIT_ISSET (hn->status, HE_STA_AUTOFIT)) HeMenuItemFit (hn);
}

void HeSetMenuItemRank (He_node *hn, int rank)
{
    OBJ_TYPE *ob;
    
    if (HeNotMenuItem(hn)) return;
    ob = hn->OBJ_UNION;
    
    ob->rank = rank;
}

void HeSetMenuItemNotifyProc (He_node *hn, He_notify_proc proc)
{
    if (HeNotMenuItem(hn)) return;
    hn->OBJ_UNION->notify_proc = proc;
}

void HeSetMenuItemSubMenu (He_node *hn, He_node *sub_menu)
{
    if (HeNotMenuItem(hn)) return;
    if (sub_menu == NULL)
        hn->OBJ_UNION->submenu_id = 0;
    else if (HeIsMenu (sub_menu))
        hn->OBJ_UNION->submenu_id = sub_menu->id;
    else HeError ("HeSetMenuItemSubMenu: menu expected in arg 2\n");
}


/*
 * Get
*/

char *HeGetMenuItemLabel (He_node *hn)
{
    if (HeNotMenuItem(hn)) return he_empty_str;
    return hn->OBJ_UNION->label ? hn->OBJ_UNION->label : he_empty_str;
}

int HeGetMenuItemRank (He_node *hn)
{
    if (HeNotMenuItem(hn)) return -1;
    return hn->OBJ_UNION->rank;
}

He_notify_proc HeGetMenuItemNotifyProc (He_node *hn)
{
    if (HeNotMenuItem(hn)) return NULL;
    return hn->OBJ_UNION->notify_proc;
}

He_node *HeGetMenuItemSubMenu (He_node *hn)
{
    if (HeNotMenuItem(hn)) return NULL;
    return HeGetNodeFromId(hn->OBJ_UNION->submenu_id);
}


/*
 *  Teste nature du noeud ; renvoie FALSE si erreur ; silencieux.
*/

int HeIsMenuItem (He_node *hn)
{
    return (hn == NULL || hn->meth != &he_meth_menu_item) ? FALSE : TRUE;
}


/*-------------------- M E T H O D E S - P R I V E E S ------------------------
 *
 * On ne teste pas si hn == NULL.
*/

/*
 * Initialisation du module ; appel par HeCreateMenuItem().
*/

void HeMenuItemAutoInit ()
{
    static int did_init = FALSE;
    
    if (did_init) return;
    
    HeMethodDefault (&he_meth_menu_item, HE_NAT_MENU_ITEM);
    he_meth_menu_item.props   = HE_PM_ACCEPT_FOCUS;
    he_meth_menu_item.destroy = HeMenuItemDestroy;
    he_meth_menu_item.fit     = HeMenuItemFit;
    he_meth_menu_item.update  = HeMenuItemUpdate;
    he_meth_menu_item.event   = HeMenuItemEvent;
    he_meth_menu_item.draw    = HeMenuItemDraw;
    
    did_init = TRUE;
}


/*
 * Destruction d'un menu_item ; appele' par HeDestroyRec().
*/

void HeMenuItemDestroy (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;

    free (ob->label);

    free (ob);
}


/*
 * Ajuste la taille du menu_item en fonction de son contenu.
*/

void HeMenuItemFit (He_node *hn)
{
    int xm, ym;
    
    if (HeNotMenuItem(hn)) return;
    
    HeMenuItemSize (hn, HE_DECO_TW, HE_DECO_TH, &xm, &ym);
    HeSetWidth (hn, xm);
    HeSetHeight(hn, ym);
    HeUpdateMark (hn, HE_UPD_DRAW);
}


/*
 * M-A-J geometrie et affichage ; appele' par HeUpdateFlush().
*/

void HeMenuItemUpdate (He_node *hn, He_upd_info *ui)
{
    HeUpdateDrawOldBg (hn, ui);

    /* Se'curite' pour que le nom ne de'passe pas le bouton */
    if (HE_BIT_ISSET (ui->mask, HE_UPD_SIZE)) {
	int x, y;

	HeMenuItemSize (hn, HE_DECO_TM, HE_DECO_TM, &x, &y);
	if (hn->xm < x) hn->xm = x;
	if (hn->ym < y) hn->ym = y;
    }

    if (HE_BIT_ISSET (ui->mask, HE_UPD_ANYDIM | HE_UPD_DRAW | HE_UPD_FOCUS)) 
	HeMenuItemDraw (hn);
}


/*
 * Methode pour gerer les evenements.
 * Renvoie -1 si erreur.
 *
 * Appele' par HeMenuEvent()
*/

int HeMenuItemEvent (He_node *hn, He_event *hev)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    int redraw = FALSE;

#ifndef DEBUG
    printf ("menu_item %s %s\n", ob->label, HeGetXEventName (hev->type)); 
#endif

    switch (hev->type) {

	case KeyPress :
            if (!HE_BIT_ISSET (hn->status, HE_STA_ACTIVE)) break;
            if (!(hev->sym == XK_Return ||
                  hev->sym == XK_KP_Enter ||
                  hev->sym == XK_ISO_Enter)) break;
            if (HE_BIT_ISSET (hn->status, HE_STA_PUSHED)) break;
	    HE_BIT_SET (hn->status, HE_STA_PUSHED);
	    redraw = TRUE;
	    break; 
            
	case ButtonPress :
            if (!HE_BIT_ISSET (hn->status, HE_STA_ACTIVE)) break;
            if (HE_BIT_ISSET (hn->status, HE_STA_PUSHED)) break;
            if (HE_BIT_ISSET (hev->xev->xbutton.state, ShiftMask))
		 HeSetFocus (hn, TRUE); 
	    else HE_BIT_SET (hn->status, HE_STA_PUSHED);
	    redraw = TRUE;
	    break; 

	case KeyRelease :
            if (!(hev->sym == XK_Return ||
                  hev->sym == XK_KP_Enter ||
                  hev->sym == XK_ISO_Enter)) break;
	case ButtonRelease :
            if (!HE_BIT_ISSET (hn->status, HE_STA_ACTIVE)) break;
            if (!HE_BIT_ISSET (hn->status, HE_STA_PUSHED)) break;
	    HE_BIT_CLR (hn->status, HE_STA_PUSHED);
	    redraw = TRUE;
	    /* Appelle la notify_proc */
	    if (ob->notify_proc != NULL)
		(ob->notify_proc)(hn);
	    break; 

	case MotionNotify :
            printf ("MotionNotify dans menu_item %d\n", hn->id);
	    break;	

	case EnterNotify : 
	    HE_BIT_SET (hn->status, HE_STA_MOUSE_IN);
	    redraw = TRUE;
	    break; 

	case LeaveNotify : 
	    HE_BIT_CLR (hn->status, HE_STA_PUSHED);
            HE_BIT_CLR (hn->status, HE_STA_MOUSE_IN);
	    redraw = TRUE;
	    break; 

	case Expose :
	    redraw = TRUE;
	    break; 

    } /* switch */

    if (redraw) HeMenuItemDraw (hn);

    return 0;
}


/*
 * Dessin du bouton.
 *
*/

void HeMenuItemDraw (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    Window win = hn->win;
    int show     = HE_BIT_ISSET (hn->status, HE_STA_SHOW),
        pushed   = HE_BIT_ISSET (hn->status, HE_STA_PUSHED),
        mouse_in = HE_BIT_ISSET (hn->status, HE_STA_MOUSE_IN),
        active   = HE_BIT_ISSET (hn->status, HE_STA_ACTIVE),
        focus    = HE_BIT_ISSET (hn->status, HE_STA_FOCUS_IN),
        survol   = active && (mouse_in || pushed),
        c_bg     = (active && pushed) ? 
                       he_col_rev_bg : he_col_bg,
        c1_ext   = (pushed) ? he_col_dark_bd : he_col_light_bd,
        c2_ext   = (pushed) ? he_col_light_bd : he_col_dark_bd,
        c1_int   = (survol) ? he_col_dark_bd : c_bg,
        c2_int   = (survol) ? he_col_light_bd : c_bg,
        text     = (ob->label && ob->label[0]);

/* printf ("draw menu_item %s\n", ob->label); */

    /* Dessin du fond */
    HeDrawBg (hn, (show) ? c_bg : he_col_bg);
    if (!show) return;

    /* Dessin bord ou focus */
    if (focus && active)
        HeDrawRect3DC (win, hn->xb, hn->yb, hn->xm-1, hn->ym-1,
	    he_black, he_black );

    /* Survole' par la souris */
    if (survol) {
	HeDrawRect3DC (win, hn->xb, hn->yb, hn->xm-1, hn->ym-1,
	    c1_ext, c2_ext);
	HeDrawRect3D (win, hn->xb+1, hn->yb+1, hn->xm-3, hn->ym-3,
	    c1_int, c2_int);
    }
    
    /* Nom */
    if (text) {
        int yb = hn->yb,
            ym = hn->ym - (yb - hn->yb);
    
        XSetForeground (he_display, he_gc_draw,
            (!active) ? he_col_light_bd :
            (pushed) ? he_col_rev_ink : he_col_ink );
        HeDrawStringCenterRect(win, he_gc_draw, he_normal_font,
            hn->xb, yb, hn->xm, ym, ob->label);
    }
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Calcule la taille xm,ym du bouton selon label
 *
*/
void HeMenuItemSize (He_node *hn, int deco_x, int deco_y, int *xm, int *ym)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    int text = (ob->label && ob->label[0]);
    
    if (text) {
	HeDimString (he_normal_font, ob->label, xm, ym);
	*xm += deco_x;
        *ym += deco_y;
    } else {
	*xm = deco_x; 
        *ym = deco_y + he_font_height; 
    }
}


/*
 *  Teste nature du noeud ; renvoie TRUE si erreur. 
*/

int HeNotMenuItem (He_node *hn)
{
    if (hn == NULL || hn->meth != &he_meth_menu_item) {
	HeError ("MenuItem expected\n");
        return TRUE;
    }
    return FALSE;
}

