/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/pan_rule.c - 22/09/1999
 *
 * Ce widget dessine un trait horizontal bicolore, utilise' par exemple
 * dans les boites de dialogue.
*/

#include <helium.h>

/* Type et union du package 
*/
#define  OBJ_TYPE	Obj_panel_rule
#define  OBJ_UNION	obj.rule

static He_method he_meth_rule;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 *  Cration d'un panel_rule.
 *  Renvoie le noeud ou NULL.
*/

He_node *HeCreateRule (He_node *owner)
{
    He_node *hn;
    OBJ_TYPE *ob;

    HeRuleAutoInit();
    
    /* On cree le noeud */
    hn = HeBaseCreate (owner, &he_meth_rule);
    if (hn == NULL) return NULL;

    /* On cree l'objet */
    ob = hn->OBJ_UNION = malloc (sizeof (OBJ_TYPE));
    if (ob == NULL) {
	HeError ("HeCreateRule: malloc error\n");
	return NULL;
    }

    /* Init des champs du noeud */    
    hn->xm = hn->ym = 0;		/* Inhibe le 1er HeUpdateDrawOldBg */
    hn->win = hn->owner->win;

    /* init des champs de l'objet */
    ob->repaint_proc = HeRuleDefaultRepaintProc;

    /* Positionne l'objet par rapport a PANEL_LAYOUT */
    HePosPanelItem (hn);

    /* Fixe taille par defaut, provoque HE_UPD_ANYDIM | HE_UPD_DRAW */
    HeRuleFit (hn);

    /* Fin succes */
    return hn;
}


/*
 * RepaintProc : attention, les dessins doivent etre faits
 * de (hn->xb, hn->yb)  (hn->xb+hn->xm-1, hn->yb+hn->ym-1)
 * car le widget ne fait pas le coupage.
*/

void HeSetRuleRepaintProc (He_node *hn, He_repaint_proc proc)
{
    if (HeNotRule(hn)) return;   
    hn->OBJ_UNION->repaint_proc = proc;
}


He_repaint_proc HeGetRuleRepaintProc (He_node *hn)
{
    if (HeNotRule(hn)) return NULL;
    return hn->OBJ_UNION->repaint_proc;
}


/*
 * RepaintProc par dfaut.
*/

void HeRuleDefaultRepaintProc (He_node *hn, Window win)
{
    HeDrawRect3D (win, hn->xb, hn->yb + (hn->ym-1)/2, hn->xm-1, 2, 
	he_col_light_bd, he_col_dark_bd); 
}


/*-------------------- M E T H O D E S - P R I V E E S ------------------------
 *
 * On ne teste pas si hn == NULL.
*/

/*
 * Initialisation du module ; appel par HeCreateRule().
*/

void HeRuleAutoInit ()
{
    static int did_init = FALSE;
    
    if (did_init) return;
    
    HeMethodDefault (&he_meth_rule, HE_NAT_RULE);
    he_meth_rule.destroy = HeRuleDestroy;
    he_meth_rule.fit     = HeRuleFit;
    he_meth_rule.update  = HeRuleUpdate;
    he_meth_rule.event   = HeRuleEvent;
    he_meth_rule.draw    = HeRuleDraw;
    
    did_init = TRUE;
}


/*
 * Destruction d'un panel_rule ; appele' par HeDestroyRec().
*/

void HeRuleDestroy (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;

    free (ob);
}


/*
 * Ajuste la taille du panel_rule en fonction de son contenu.
*/

void HeRuleFit (He_node *hn)
{
    if (HeNotRule(hn)) return;
    
    HeSetWidth (hn, 100);
    HeSetHeight(hn, he_font_height + HE_DECO_TH);
    HeUpdateMark (hn, HE_UPD_DRAW);
}


/*
 * M-A-J geometrie et affichage ; appele' par HeUpdateFlush().
*/

void HeRuleUpdate (He_node *hn, He_upd_info *ui)
{
    /* OBJ_TYPE *ob = hn->OBJ_UNION; */
    
    HeUpdateDrawOldBg (hn, ui);

    /* Se'curite' taille */
    if (HE_BIT_ISSET (ui->mask, HE_UPD_DIMY)) {
	if (hn->ym < 2) hn->ym = 2;
    }

    if (HE_BIT_ISSET (ui->mask, HE_UPD_ANYDIM | HE_UPD_DRAW | HE_UPD_FOCUS)) 
	HeRuleDraw (hn);
}


/*
 * Methode pour gerer les evenements.
 * Renvoie -1 si erreur.
 *
 * Appele' par HePanelEvent()
*/

int HeRuleEvent (He_node *hn, He_event *hev)
{

    switch (hev->type) {

	case Expose :
	    HeRuleDraw (hn);
	    break; 

    } /* switch */

    return 0;
}


/*
 * Dessin du widget.
 *
*/

void HeRuleDraw (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    int show = HE_BIT_ISSET (hn->status, HE_STA_SHOW);

    /* Dessin du fond */
    HeDrawBg (hn, he_col_bg);
    if (!show) return;

    /* Dessin rule */
    if (ob->repaint_proc != NULL)
	(ob->repaint_proc)(hn, hn->win);
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 *  Teste nature du noeud ; renvoie TRUE si erreur. 
*/

int HeNotRule (He_node *hn)
{
    if (hn == NULL || hn->meth != &he_meth_rule) {
	HeError ("Rule expected\n");
        return TRUE;
    }
    return FALSE;
}

