/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/pan_text.c - 19/06/2000
 *
 * 
*/

#include <helium.h>

/* Type et union du package 
*/
#define  OBJ_TYPE	Obj_panel_text
#define  OBJ_UNION	obj.text

static He_method he_meth_text;


/* Masques pour hn->ob->type */
enum {
	HE_TT_NULL		= 0L,
	HE_TT_COMPLETION	= 1L<<0 ,
	HE_TT_HISTORY		= 1L<<1 ,
	HE_TT_INTEGER		= 1L<<2 ,
	HE_TT_PASSWD		= 1L<<3 ,
};

static Time he_time_clic = 0;
static int he_nb_clic = 0, he_sel1 = 0, he_sel2 = 0, he_sel_deb = 0,
           he_move_sel = FALSE, he_x_sel;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 *  Cration d'un panel_text.
 *  Renvoie le noeud ou NULL.
*/

He_node *HeCreateText (He_node *owner)
{
    He_node *hn;
    OBJ_TYPE *ob;

    HeTextAutoInit();
    
    /* On cree le noeud */
    hn = HeBaseCreate (owner, &he_meth_text);
    if (hn == NULL) return NULL;

    /* On cree l'objet */
    ob = hn->OBJ_UNION = malloc (sizeof (OBJ_TYPE));
    if (ob == NULL) {
	HeError ("HeCreateText: malloc error\n");
	return NULL;
    }

    /* Init des champs du noeud */    
    hn->xm = hn->ym = 0;		/* Inhibe le 1er HeUpdateDrawOldBg */
    hn->win = hn->owner->win;
    HE_BIT_SET (hn->status, HE_STA_INSERT_MODE);

    /* init des champs de l'objet */
    ob->value = NULL;
    ob->type = HE_TT_NULL;                  
    ob->str_len = 0;
    ob->vis_len = 32;
    ob->offset = 0;              
    ob->point = 0;              
    ob->kbd_proc = NULL;
    ob->notify_proc = NULL;
    HeTipInitInfo (&ob->tip_info);
    
    /* Positionne l'objet par rapport a PANEL_LAYOUT */
    HePosPanelItem (hn);

    /* Fixe taille par defaut, provoque HE_UPD_ANYDIM | HE_UPD_DRAW */
    HeTextFit (hn);

    /* Fin succes */
    return hn;
}


/*
 * Set - redessin systematique
*/

void HeSetTextValue (He_node *hn, const char *value)
{
    OBJ_TYPE *ob;
    
    if (HeNotText(hn)) return; ob = hn->OBJ_UNION;
    
    free (ob->value);
    ob->value = (value == NULL) ? NULL : strdup(value);
    ob->str_len = HeStrLen (ob->value);
    
    HeTextBounds (hn);
    HeUpdateMark (hn, HE_UPD_DRAW);
    
    /* Annule la selection */
    if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) HeClearSelection ();
}

void HeSetTextVisibleLen (He_node *hn, int len)
{
    if (HeNotText(hn)) return;
    hn->OBJ_UNION->vis_len = len;
    
    HeTextFit (hn);
}


/*
 * Set - redessin uniquement si necessaire
*/

void HeSetTextPoint (He_node *hn, int point)
{
    OBJ_TYPE *ob;
    int old_point, old_offset;
    
    if (HeNotText(hn)) return; ob = hn->OBJ_UNION;
    
    old_point = ob->point; old_offset = ob->offset; ob->point = point; 
    HeTextBounds (hn);
    if (old_point != ob->point || old_offset != ob->offset)
	HeUpdateMark (hn, HE_UPD_DRAW);
}

void HeSetTextOffset (He_node *hn, int offset)
{
    OBJ_TYPE *ob;
    int old_offset;
    
    if (HeNotText(hn)) return; ob = hn->OBJ_UNION;
    
    old_offset = ob->offset; ob->offset = offset; 
    HeTextBounds (hn);
    if (old_offset != ob->offset)
	HeUpdateMark (hn, HE_UPD_DRAW);
}

void HeSetTextInsertMode (He_node *hn, int insert_mode)
{
    if (HeNotText(hn)) return; 
    if (insert_mode)
	 HE_BIT_SET (hn->status, HE_STA_INSERT_MODE);
    else HE_BIT_CLR (hn->status, HE_STA_INSERT_MODE);
    HeUpdateMark (hn, HE_UPD_DRAW);
}


/*
 * Set sans redessin
*/

void HeSetTextKbdProc (He_node *hn, He_kbd_proc proc)
{
    if (HeNotText(hn)) return; 
    hn->OBJ_UNION->kbd_proc = proc;
}

void HeSetTextNotifyProc (He_node *hn, He_notify_proc proc)
{
    if (HeNotText(hn)) return; 
    hn->OBJ_UNION->notify_proc = proc;
    
}

void HeSetTextCompletion (He_node *hn, int completion)
{
    if (HeNotText(hn)) return; 
    if (completion)
	 HE_BIT_SET (hn->OBJ_UNION->type, HE_TT_COMPLETION);
    else HE_BIT_CLR (hn->OBJ_UNION->type, HE_TT_COMPLETION);
}


/*
 * Get - ATTENTION, le retour de HeGetTextValue est TRES VOLATILE
*/

char *HeGetTextValue (He_node *hn)
{
    if (HeNotText(hn)) return he_empty_str;
    return hn->OBJ_UNION->value ?
	hn->OBJ_UNION->value : he_empty_str;
}

int HeGetTextLen (He_node *hn)
{
    if (HeNotText(hn)) return 0;
    return hn->OBJ_UNION->str_len;
}

int HeGetTextVisibleLen (He_node *hn)
{
    if (HeNotText(hn)) return 0;
    return hn->OBJ_UNION->vis_len;
}

int HeGetTextPoint (He_node *hn)
{
    if (HeNotText(hn)) return 0;
    return hn->OBJ_UNION->point;
}

int HeGetTextOffset (He_node *hn)
{
    if (HeNotText(hn)) return 0;
    return hn->OBJ_UNION->offset;
}

int HeGetTextInsertMode (He_node *hn)
{
    if (HeNotText(hn)) return FALSE; 
    return HE_BIT_ISSET (hn->status, HE_STA_INSERT_MODE);  
}

He_kbd_proc HeGetTextKbdProc (He_node *hn)
{
    if (HeNotText(hn)) return NULL; 
    return hn->OBJ_UNION->kbd_proc;
}

He_notify_proc HeGetTextNotifyProc (He_node *hn)
{
    if (HeNotText(hn)) return NULL; 
    return hn->OBJ_UNION->notify_proc;
}

int HeGetTextCompletion (He_node *hn)
{
    if (HeNotText(hn)) return FALSE; 
    return HE_BIT_ISSET (hn->OBJ_UNION->type, HE_TT_COMPLETION);  
}


/*
 * Operations sur les string, au niveau du point d'insertion ob->point.
 * Met a jour la selection eventuelle : decalage, anulation (mais pas
 * remplacement : c'est HeTextInput() qui s'en charge).
 * Renvoie 0 succes, -1 erreur.
*/

int HeTextInsert (He_node *hn, const char *s, int len)
{
    OBJ_TYPE *ob;
    int i;
    char *tmp;
    
    if (HeNotText(hn)) return -1; ob = hn->OBJ_UNION;
    
    if (s == NULL || s[0] == 0 || len <= 0) return -1;
    
    tmp = malloc (ob->str_len+len+1);
    if (tmp == NULL) return -1;
    
    for (i = 0; i < ob->point; i++) tmp[i] = ob->value[i];
    for (i = 0; i < len; i++) tmp[ob->point+i] = s[i];
    for (i = ob->point; i < ob->str_len; i++) tmp[len+i] = ob->value[i];
    tmp[ob->str_len+len] = 0;

    free (ob->value);
    ob->value = tmp;
    ob->str_len = ob->str_len+len;
    
    /* Met a jour la selection eventuelle */
    if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) {
	if (ob->point <= he_sel1) {
            he_sel_deb += len; he_sel1 += len; he_sel2 += len;
        } else if (ob->point < he_sel2)
	    HeClearSelection (hn);
        /* sinon on laisse tel quel */
    }

    /* Change le point d'insertion */
    ob->point += len;
    HeTextBounds (hn);
    HeUpdateMark (hn, HE_UPD_DRAW);
    
    return 0;
}


int HeTextOvertype (He_node *hn, const char *s, int len)
{
    OBJ_TYPE *ob;
    int  i;
    
    if (HeNotText(hn)) return -1; ob = hn->OBJ_UNION;
    
    if (s == NULL || s[0] == 0 || len <= 0) return -1;
    
    if (ob->point+len <= ob->str_len) {
    
	for (i = 0; i < len; i++) ob->value[i+ob->point] = s[i];
    
    } else {
	char *tmp;
        
	tmp = malloc (ob->point+len+1);
	if (tmp == NULL) return -1;
        
	for (i = 0; i < ob->str_len; i++) tmp[i] = ob->value[i];
	for (i = 0; i < len; i++) tmp[i+ob->point] = s[i];
        tmp[ob->point+len] = 0;

	free (ob->value);
	ob->value = tmp;
        ob->str_len = ob->point+len;
    }
    
    /* Met a jour la selection eventuelle */
    if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) {
	if (ob->point+len > he_sel1 && he_sel2 > ob->point)
	    HeClearSelection (hn);
        /* sinon on laisse tel quel */
    }

    /* Change le point d'insertion */
    ob->point += len;
    HeTextBounds (hn);
    HeUpdateMark (hn, HE_UPD_DRAW);
    
    return 0;
}


int HeTextDelete (He_node *hn, int len)
{
    OBJ_TYPE *ob;
    int i;
    char *tmp;
    
    if (HeNotText(hn)) return -1; ob = hn->OBJ_UNION;
    
    if (len <= 0 || ob->point >= ob->str_len) return -1;
    if (len > ob->str_len - ob->point) len = ob->str_len - ob->point;
    
    tmp = malloc (ob->str_len-len+1);
    if (tmp == NULL) return -1;
    
    for (i = 0; i < ob->point; i++) tmp[i] = ob->value[i];
    for (i = ob->point+len; i < ob->str_len; i++) tmp[i-len] = ob->value[i];
    tmp[ob->str_len-len] = 0;

    free (ob->value);
    ob->value = tmp;
    ob->str_len = ob->str_len-len;
    
    /* Met a jour la selection eventuelle */
    if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) {
	if (ob->point+len > he_sel1 && he_sel2 > ob->point)
	    HeClearSelection (hn);
        else if (ob->point <= he_sel1) {
            he_sel_deb -= len; he_sel1 -= len; he_sel2 -= len;
        } 
        /* sinon on laisse tel quel */
    }

    HeTextBounds (hn);
    HeUpdateMark (hn, HE_UPD_DRAW);
    
    return 0;
}


/*
 * On insere le texte en mode Insert ou Overtype,
 * en remplacant la selection si ob->point est dedans.
 *
 * ATTENTION : si s pointe dans la selection et qu'on la supprime, alors
 * s est altere' !! Dans ce cas il faut l'appel avec delete_sel = FALSE.
*/

int HeTextInput (He_node *hn, char *s, int len, int delete_sel)
{
    OBJ_TYPE *ob;
    
    if (HeNotText(hn)) return -1; ob = hn->OBJ_UNION;
    
    /* Si du texte est selectionne, on l'efface si ob->point est dedans */
    if ( delete_sel && HE_BIT_ISSET (hn->status, HE_STA_SELECTION) &&
         he_sel1 <= ob->point && ob->point <= he_sel2)
    {
	HeSetTextPoint (hn, he_sel1);
	HeTextDelete (hn, he_sel2 - he_sel1);
    }

    /* On insere le texte en mode Insert ou Overtype */
    if (HE_BIT_ISSET (hn->status, HE_STA_INSERT_MODE))
         return HeTextInsert   (hn, s, len);
    else return HeTextOvertype (hn, s, len);
}


/*
 * Selection
 *
 * Renvoie 0 succes, -1 erreur.
*/

int HeTextSelIndex (He_node *hn, int sel1, int sel2)
{
    OBJ_TYPE *ob;
    
    if (HeNotText(hn)) return -1; ob = hn->OBJ_UNION;
    
    if (sel1 > sel2 || sel1 < 0 || sel2 > ob->str_len || 
	ob->str_len == 0) return -1;
    
    HeSetSelection (hn); 
    he_sel_deb = he_sel1 = sel1; he_sel2 = sel2;
    HeUpdateMark (hn, HE_UPD_DRAW);
    
    return 0;
}


int HeTextSelWord (He_node *hn)
{
    OBJ_TYPE *ob;
    int i, (*fct_test)(int c) = HeIsAlnum;
    
    if (HeNotText(hn)) return -1; ob = hn->OBJ_UNION;
    
    HeSetSelection (hn); he_sel_deb = he_sel1 = he_sel2 = 0;
    if (ob->str_len == 0) return -1;
    
    if (HeIsAlnum(ob->value[ob->point])) fct_test = HeIsAlnum;
    else if (HeIsSpace(ob->value[ob->point])) fct_test = HeIsSpace;
    else if (HeIsPunct(ob->value[ob->point])) fct_test = HeIsPunct;
    else fct_test = HeIsCntrl;
    
    for (i = ob->point; i > 0; i--)
	if (!fct_test(ob->value[i-1])) break;
    he_sel_deb = he_sel1 = i; 
    for (i = ob->point; i < ob->str_len; i++)
	if (!fct_test(ob->value[i])) break;
    he_sel2 = i; 
    HeUpdateMark (hn, HE_UPD_DRAW);
    
    return 0;
}


int HeTextSelLine (He_node *hn)
{
    OBJ_TYPE *ob;
    
    if (HeNotText(hn)) return -1; ob = hn->OBJ_UNION;
    
    HeSetSelection (hn);
    he_sel_deb = he_sel1 = 0; he_sel2 = ob->str_len;
    HeUpdateMark (hn, HE_UPD_DRAW);
    
    return 0;
}


int HeTextMoveSel (He_node *hn, int dec)
{
    OBJ_TYPE *ob;
    int i, depla; char svg[101];
    
    if (HeNotText(hn) || !HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) return -1; 
    ob = hn->OBJ_UNION;
    if ( ob->str_len == 0 || he_sel1 >= he_sel2 || dec == 0 ) return -1;
    
    depla = dec;  /* dec vas etre reborne' ; depla permettra d'amener 
                     ob->point jusqu'aux bords */
    if (dec < 0) {
        if (dec < -100) dec = -100;
        if (he_sel1 + dec < 0) dec = - he_sel1;
    } else {
        if (dec > 100) dec = 100;
	if (he_sel2 + dec > ob->str_len) dec = ob->str_len - he_sel2;
    }
    
    if (dec < 0) {
        for (i = 0; i < -dec; i++)
            svg[i] = ob->value[he_sel1+dec+i];
        for (i = he_sel1; i < he_sel2; i++)
            ob->value[i+dec] = ob->value[i];
        for (i = 0; i < -dec; i++)
            ob->value[he_sel2+dec+i] = svg[i];
    } else if (dec > 0) {
        for (i = 0; i < dec; i++)
            svg[i] = ob->value[he_sel2+i];
        for (i = he_sel2-1; i >= he_sel1; i--)
            ob->value[i+dec] = ob->value[i];
        for (i = 0; i < dec; i++)
            ob->value[he_sel1+i] = svg[i];
    }
    he_sel1 += dec; he_sel2 += dec; he_sel_deb +=  dec;
    
    /* N'affiche que si necessaire */
    HeMoveTextPoint (hn, depla);
    
    return 0;
}


int HeTextPasteSel (He_node *hn)
{
    OBJ_TYPE *ob;
    int interne = FALSE, len = 0;
    char *s = NULL;
    He_node *hop;
    
    if (HeNotText(hn)) return -1; ob = hn->OBJ_UNION;

    if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) {
    
	if ( ob->value != NULL && 0 <= he_sel1 && 
	     he_sel1 <= he_sel2 && he_sel2 <= ob->str_len)
        {
            s = ob->value + he_sel1;
            len = he_sel2 - he_sel1;
            interne = TRUE;
        }
        
    } else if ((hop = HeGetSelection ()) != NULL) {
        
	if (hop->meth == &he_meth_text) {
        
            Obj_panel_text *hop_ob = hop->obj.text;
            
            if ( hop_ob->value != NULL && 0 <= he_sel1 && 
                 he_sel1 <= he_sel2 && he_sel2 <= hop_ob->str_len)
            {
                s = hop_ob->value + he_sel1;
                len = he_sel2 - he_sel1;
                interne = TRUE;
            }
        }
    }
        
    if (interne)
        /* On insere le texte sans remplacer la selection */
        return HeTextInput (hn, s, len, FALSE);
    else {
    	/* Demande la selection PRIMARY au serveur X */
        HeImportSelection (hn, XA_PRIMARY, XA_STRING, he_atom_sel0);
    }

    return 0;
}


int HeTextPasteClip (He_node *hn)
{
    char *s = NULL;
    int len = 0, interne = FALSE;
    
    if (HeNotText(hn)) return -1;

    /* Tester ici si on possede le clipboard en interne */
    HeGetClipboardString (&s, &len, &interne);

    if (interne) return HeTextInput (hn, s, len, TRUE);
    else {
	/* Demande la selection CLIPBOARD au serveur X */
	HeImportSelection (hn, he_atom_clipboard, XA_STRING, he_atom_sel0);
    }
    
    return 0;
}


int HeTextCutClip (He_node *hn)
{
    if (HeNotText(hn)) return -1;

    /* Verifie qu'il y a du texte selectionne,
     * puis copie le texte selectionne.
    */
    if (HeTextCopyClip (hn) < 0) return -1;
    
    /* On efface la selection */
    HeSetTextPoint (hn, he_sel1);
    HeTextDelete (hn, he_sel2 - he_sel1);
    HeClearSelection ();
    
    return 0;
}


int HeTextCopyClip (He_node *hn)
{
    OBJ_TYPE *ob;
    
    if (HeNotText(hn)) return -1; ob = hn->OBJ_UNION;

    if (!HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) return -1;
    
    /* On copie la selection dans le clipboard */
    if ( ob->value != NULL && 0 <= he_sel1 && 
	 he_sel1 <= he_sel2 && he_sel2 <= ob->str_len)
    {
	HeSetClipboardString (ob->value + he_sel1, he_sel2 - he_sel1);
        return 0;
    }
    
    return -1;
}


/*
 *  Tests isalpha() : variante. 
*/

int HeIsAlnum (int c)
{
    return (isalnum(c) || c == '_' || c < 0 || c > 127);
}

int HeIsSpace (int c)
{
    return isspace(c);
}

int HeIsPunct (int c)
{
    return (ispunct(c) && c != '_');
}

int HeIsCntrl (int c)
{
    return (!(HeIsAlnum(c) || HeIsSpace(c) || HeIsPunct(c)));
}


/*
 * Macros utiles
*/

void HeMoveTextPoint (He_node *hn, int move)
{
    HeSetTextPoint (hn, HeGetTextPoint(hn) + move);
}

void HeMoveTextOffset (He_node *hn, int move)
{
    HeSetTextOffset (hn, HeGetTextOffset(hn) + move);
}


/*-------------------- M E T H O D E S - P R I V E E S ------------------------
 *
 * On ne teste pas si hn == NULL.
*/

/*
 * Initialisation du module ; appel par HeCreateText().
*/

void HeTextAutoInit ()
{
    static int did_init = FALSE;
    
    if (did_init) return;
    
    HeMethodDefault (&he_meth_text, HE_NAT_TEXT);
    he_meth_text.props      = HE_PM_ACCEPT_FOCUS | HE_PM_SEND_TAB;
    he_meth_text.destroy    = HeTextDestroy;
    he_meth_text.fit        = HeTextFit;
    he_meth_text.update     = HeTextUpdate;
    he_meth_text.event      = HeTextEvent;
    he_meth_text.draw       = HeTextDraw;
    he_meth_text.sel_export = HeTextSelExport;
    he_meth_text.sel_import = HeTextSelImport;
    he_meth_text.tip        = HeTextStoreTip;
    
    did_init = TRUE;
}


/*
 * Destruction d'un panel_text ; appele' par HeDestroyRec().
*/

void HeTextDestroy (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;

    free (ob->value);
    HeTipDestroyInfo (&ob->tip_info);

    free (ob);
}


/*
 * Ajuste la taille du panel_text en fonction de son contenu.
 * Provoque un Update et un appel a HeTextBounds.
*/

void HeTextFit (He_node *hn)
{
    OBJ_TYPE *ob;
    
    if (HeNotText(hn)) return; ob = hn->OBJ_UNION;
    
    HeSetWidth (hn, he_font_width * ob->vis_len + HE_DECO_TW); 
    HeSetHeight(hn, he_font_height + HE_DECO_TH);
    HeUpdateMark (hn, HE_UPD_DRAW);
}


/*
 * M-A-J geometrie et affichage ; appele' par HeUpdateFlush().
*/

void HeTextUpdate (He_node *hn, He_upd_info *ui)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    
    HeUpdateDrawOldBg (hn, ui);

    if (HE_BIT_ISSET (ui->mask, HE_UPD_SIZE)) {
	/* Taille minimale */
        if (hn->xm < HE_DECO_TW)
            hn->xm = HE_DECO_TW;
	if (hn->ym < he_font_height + HE_DECO_TM)
            hn->ym = he_font_height + HE_DECO_TM;

	/* Recalcul vis_len */
        ob->vis_len = (hn->xm - HE_DECO_TW) / he_font_width;
        if (ob->vis_len < 0) ob->vis_len = 0;
        
	HeTextBounds (hn);
    }
    
    if (HE_BIT_ISSET (ui->mask, HE_UPD_ANYDIM | HE_UPD_DRAW | HE_UPD_FOCUS)) 
	HeTextDraw (hn);
}


/*
 * Methode pour gerer les evenements.
 * Renvoie -1 si erreur.
 *
 * Appele' par HePanelEvent()
*/

int HeTextEvent (He_node *hn, He_event *hev)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    int redraw = FALSE;

    HeTipEvent (hn, hev);
    
    if (!HE_BIT_ISSET (hn->status, HE_STA_ACTIVE)) {
    
      switch (hev->type) {
	case EnterNotify : 
	    HE_BIT_SET (hn->status, HE_STA_MOUSE_IN);
	    redraw = TRUE;
	    break; 

	case LeaveNotify : 
	    HE_BIT_CLR (hn->status, HE_STA_MOUSE_IN);
	    redraw = TRUE;
	    break; 

	case Expose :
	    redraw = TRUE;
	    break; 
      }
    
    } else switch (hev->type) {

	case KeyPress :
	case KeyRelease :
	    /* Appelle kbd_proc ; intercepte l'evt si elle renvoie FALSE */
	    if (ob->kbd_proc != NULL)
		if ((ob->kbd_proc)(hn, hev) == FALSE) break;
                
            if (hev->type == KeyPress)
		HeTextKeyPress (hn, hev);
            break; 

	case EnterNotify : 
	    HE_BIT_SET (hn->status, HE_STA_MOUSE_IN);
	    redraw = TRUE;
	    break; 

	case LeaveNotify : 
	    HE_BIT_CLR (hn->status, HE_STA_MOUSE_IN);
	    redraw = TRUE;
	    break; 

	case Expose :
	    redraw = TRUE;
	    break; 

	case ButtonPress :
	    if (hev->sb == 1) {
                
                int shift = HE_BIT_ISSET (hev->xev->xbutton.state, ShiftMask);
                
                if (shift)
                     he_nb_clic = 0;        /* shift + clic */
                else switch (he_nb_clic) {  /* clic simple, double ou triple */
                    case 0 :
                      he_nb_clic = 1;
                      break;
                    case 1 :
                      if (hev->time - he_time_clic < he_multiclic_timeout)
                           he_nb_clic = 2;
                      else he_nb_clic = 1;
                      break;
                    case 2 :
                      if (hev->time - he_time_clic < he_multiclic_timeout)
                           he_nb_clic = 3;
                      else he_nb_clic = 1;
                      break;
                }
                he_time_clic = hev->time;
                
                /* Action suivant le clic */
                switch (he_nb_clic) {
                  case 0 : /* shift + clic : selection / milieu */
                    HeSetFocus (hn, TRUE);
                    if (!HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) {
                        HeSetSelection (hn);
                        he_sel_deb = ob->point;
                        HeSetTextPoint (hn, HeTextSymClic (hn, hev->sx));
                        HeTextSel1Sel2 (ob->point);
                    } else {
                        HeSetTextPoint (hn, HeTextSymClic (hn, hev->sx));
                        redraw = TRUE;  /* au cas ou ob->point inchange' */
			if (ob->point < (he_sel1 + he_sel2) / 2)
                             { he_sel_deb = he_sel2; he_sel1 = ob->point; }
                        else { he_sel_deb = he_sel1; he_sel2 = ob->point; }
                    }
                    break;
                  case 1 : /* Simple clic : on place le point d'insertion */
                    HeSetFocus (hn, TRUE);
                    if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION))
                        HeClearSelection ();
                    HeSetTextPoint (hn, HeTextSymClic (hn, hev->sx));
                    break;
                  case 2 : /* Double clic : on selectionne le mot */
                    HeTextSelWord (hn);
                    break;
                  case 3 : /* Triple clic : on selectionne toute la ligne */
                    HeTextSelLine (hn);
                    break;
                }
                
            } else if (hev->sb == 2) {
            
		int x0 = HeTextCalcX0 (hn),
                    xa = x0 + HeTextSymPos (hn, he_sel1) * he_font_width,
		    xb = x0 + HeTextSymPos (hn, he_sel2) * he_font_width;
                    
		he_move_sel = FALSE;
                HeSetFocus (hn, TRUE);
                HeSetTextPoint (hn, HeTextSymClic (hn, hev->sx));
                              
                if ( HE_BIT_ISSET (hn->status, HE_STA_SELECTION) && 
                     xa <= hev->sx && hev->sx < xb )
                {
                    /* Debut deplacement selection */
                    he_move_sel = TRUE; he_x_sel = hev->sx;
                                        
                } else {
                
                    /* Paste */
                    HeTextPasteSel (hn);
                }
              
            } else if (hev->sb == 3) {
            
            } break; 
            
  	case MotionNotify :
            if (hev->sb == 1) {
                int x0, x1, old_point;
                
                /* Si on est dans un double ou triple clic on ne fait rien */
                if (he_nb_clic > 1) break;
                
                x0 = HeTextCalcX0 (hn);
                x1 = x0 + ob->vis_len * he_font_width;
                
                old_point = ob->point;
                if (hev->sx < x0) HeMoveTextPoint (hn, -1);
                else if (hev->sx > x1) HeMoveTextPoint (hn, 1);
                else HeSetTextPoint (hn, HeTextSymClic (hn, hev->sx));
                
                if (old_point != ob->point) {
        	    if (!HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) {
                        /* appele' que si le deplacement est significatif */
                        HeSetSelection (hn);
                        he_sel_deb = old_point;
                    }
                    HeTextSel1Sel2 (ob->point);
                }
                
            } else if (hev->sb == 2) {
            
        	if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION) && he_move_sel)
                {
                    int x0, x1, depla = 0;
                   
                    x0 = HeTextCalcX0 (hn);
                    x1 = x0 + ob->vis_len * he_font_width;
                    
                    if (hev->sx < x0) depla = -1; 
                    else if (hev->sx > x1) depla = 1;
                    else {
                        int dx = hev->sx - he_x_sel,
                            p  = (dx < 0) ? he_sel1 : he_sel2;
                        depla = HeTextSymDepla (hn, &dx, p);
                        if (depla != 0) he_x_sel = hev->sx - dx;
                    }
                    HeTextMoveSel (hn, depla);                    
                }
                
            } else if (hev->sb == 3) {
            
            } break;	

  	case ButtonRelease :
            if (hev->sb == 1) {
            
                /* Fin du triple clic */
                if (he_nb_clic == 3) he_nb_clic = 0;
            
                if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) {
                
                    /* L'appli devient proprietaire de l'atome PRIMARY */
                    /* HeSetPrimaryOwner (hev->time); */
                }
            
            } else if (hev->sb == 2) {
            
		/* Fin deplacement selection */
                he_move_sel = FALSE;
                
            } else if (hev->sb == 3) {
            
            } break;	

    } /* switch */

    if (redraw) HeTextDraw (hn);

    return 0;
}


/*
 * Dessin du text.
 *
*/

void HeTextDraw (He_node *hn)
{
    /* Obj_panel *ob_pan = hn->owner->obj.panel; */
    OBJ_TYPE *ob = hn->OBJ_UNION;
    Window win = hn->win;
    int show        = HE_BIT_ISSET (hn->status, HE_STA_SHOW),
        mouse_in    = HE_BIT_ISSET (hn->status, HE_STA_MOUSE_IN),
        active      = HE_BIT_ISSET (hn->status, HE_STA_ACTIVE),
        focus_own   = HE_BIT_ISSET (hn->status, HE_STA_FOCUS_OWN),
        focus_in    = HE_BIT_ISSET (hn->status, HE_STA_FOCUS_IN),
        selec       = HE_BIT_ISSET (hn->status, HE_STA_SELECTION),
        x0, y0, c1, c2, c3, c4;
        
    /* Dessin du fond */
    HeDrawBg (hn, (show) ? he_white : he_col_bg);
    if (!show) return;
    
    /* Calcul coordonnees du coin en haut a gauche du texte */
    x0 = HeTextCalcX0 (hn);
    y0 = HeTextCalcY0 (hn);
         
    /* Fond de la selection */
    if (active && selec) {
        XSetForeground (he_display, he_gc_draw, he_col_sel);
        HeTextDrawSelBg (hn, x0, y0);
    }
    
    /* Cadre */
    if (active && focus_in) { 
        c1 = he_black; c2 = he_black; 
        c3 = he_col_light_bd; c4 = he_col_bg;
    } else if (active && mouse_in) { 
	c1 = he_col_dark_bd; c2 = he_col_light_bd; 
        c3 = he_col_bg; c4 = he_col_dark_bd;
    } else {
	c1 = he_col_dark_bd; c2 = he_col_light_bd; 
        c3 = he_col_light_bd; c4 = he_col_bg;
    }
    HeDrawRect3D (win, hn->xb, hn->yb, hn->xm-1, hn->ym-1, c1, c2);
    HeDrawRect3D (win, hn->xb+1, hn->yb+1, hn->xm-3, hn->ym-3, c3, c4);
    
    /* Texte */
    XSetForeground (he_display, he_gc_draw, 
        (!active) ? he_col_bg : he_col_ink );
    HeTextSymDraw (hn, x0, y0);
        
    /* Curseur */
    if (active && focus_own) {
	int x_point = x0-1 + HeTextSymPos (hn, ob->point) * he_font_width;
	XSetForeground (he_display, he_gc_draw, he_black);
	HeTextDrawCursor (hn, x_point, y0);
    }
}


/*
 * Mthode renvoyant &tip.
*/
He_tip_info *HeTextStoreTip (He_node *hn)
{
    return &hn->OBJ_UNION->tip_info;
}


/*
 * Une autre appli X demande la selection.
 * Renvoie TRUE si on a pu repondre.
 *
 * Appele par HeXSelRequest().
*/

int HeTextSelExport (He_node *hn, Atom sel_type, Atom target,
	Atom *type, Uchar **value, int *length, int *format, int *free_value)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    
    if (sel_type != XA_PRIMARY) return FALSE;
    
    /* 
     * On nous demande le texte selectionne'
    */
    if (target == XA_STRING) {
    
        *type = XA_STRING;
        if ( ob->value != NULL && 0 <= he_sel1 && he_sel1 <= he_sel2 &&
             he_sel2 <= ob->str_len)
        { 
            *value = (Uchar *) (ob->value + he_sel1); 
            *length = he_sel2 - he_sel1; 
        }
        else { 
            *value = (Uchar *) he_empty_str; 
            *length = 0;
        }
        *format = 8;
        *free_value = FALSE;
        return TRUE;
    }
    
    /*
     * On nous demande la liste des targets qu'on peut traiter
    */
    if (target == he_atom_targets) {
    
    /*  RQ : ne plait pas a cc de SGI :
        Atom liste[] = { he_atom_targets, XA_STRING };
        int taille = sizeof(liste) / sizeof(Atom);
    */  
        Atom liste[2]; int taille = 2;
        liste[0] = he_atom_targets; liste[1] = XA_STRING;
        
        *type = XA_ATOM;
        *value = (Uchar *) liste;
        *length = taille;
        *format = 32;
        *free_value = FALSE;
        return TRUE;
    }
    
    return FALSE;
}


/*
 * On recoit une selection X demandee a une autre appli par HeImportSelection().
 *
 * RQ : - sel_type == XA_PRIMARY ou "CLIPBOARD" 
 *      - on peut se servir de he_import_* comme variables d'etat
 *
 * Appele par HeXSelNotify().
*/

void HeTextSelImport (He_node *hn, Atom sel_type, Atom target, Atom prop,
	Atom type, int format, Ulong nitems, Uchar *data, int *free_data)
{
        
    /* On liberera data apres l'appel */
    *free_data = TRUE;
    
    if (target == XA_STRING) {
        if (prop != None && type == XA_STRING && format == 8) {
        
            HeTextInput (hn, (char *) data, nitems, TRUE);
        }
         
#ifdef DEBUG
	/* On demande la liste des targets pour le meme sel_type */
        HeImportSelection (hn, sel_type, he_atom_targets, he_atom_sel0);
#endif
        return;
    }
    
    if (target == he_atom_targets) {
        if (prop != None && type == XA_ATOM && format == 32) {
	    /* on a la liste des targets acceptes ;
               pour le moment on n'en fait rien */
        }
        return;
    }
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Calcul coordonnees du coin en haut a gauche du texte.
*/

int HeTextCalcX0 (He_node *hn)
{
    return hn->xb + (hn->xm - hn->OBJ_UNION->vis_len * he_font_width )/2;
}

int HeTextCalcY0 (He_node *hn)
{
    return hn->yb + (hn->ym - he_font_height)/2;
}


/*
 * Dessin du curseur ; x,y sont les coord du milieu dessus du curseur.
*/

void HeTextDrawCursor (He_node *hn, int x0, int y0)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    Window win = hn->win;
    int x1, y1, x2, y2;
    
    x1 = x0 - 3;
    x2 = x0 + 3;
    y1 = y0;
    y2 = y0 + he_font_height - 1;
    
    if (!HE_BIT_ISSET (hn->status, HE_STA_FOCUS_IN)) {  /* dessin "^" */
    
        XDrawLine (he_display, win, he_gc_draw, x1, y2, x0, y2-3);
        XDrawLine (he_display, win, he_gc_draw, x0, y2-3, x2, y2);     
        XDrawLine (he_display, win, he_gc_draw, x1+1, y2-2, x0, y2-4);
        XDrawLine (he_display, win, he_gc_draw, x0, y2-4, x2-1, y2-2);     
        
    } else if (HE_BIT_ISSET (hn->status, HE_STA_INSERT_MODE)) {  /* dessin "I" */
    
        XDrawLine (he_display, win, he_gc_draw, x0, y1, x0, y2);
        XDrawLine (he_display, win, he_gc_draw, x1, y1, x2, y1);
        XDrawLine (he_display, win, he_gc_draw, x1, y2, x2, y2);
        
    } else { /* mode Overtype, dessin "[]" */
    
        int wmin = hn->xb + hn->xm - x0 - 1,
            wsym = he_font_width * 
                    HeTextSymLen (ob->value ? ob->value[ob->point] : ' ');
        if (wmin > wsym) wmin = wsym;
        XDrawRectangle (he_display, win, he_gc_draw,
                x0, y1, wmin, he_font_height-1);
    }
}


/*
 * Dessin du fond de la selection
*/

void HeTextDrawSelBg (He_node *hn, int x0, int y0)
{
    int xa, xb;
    
    if (he_sel1 == he_sel2) return;
    xa = x0 + HeTextSymPos (hn, he_sel1) * he_font_width;
    xb = x0 + HeTextSymPos (hn, he_sel2) * he_font_width;
    if (xa == xb) return;
    XFillRectangle (he_display, hn->win, he_gc_draw,
        xa, y0, xb-xa, he_font_height);
}


/*
 * Calcul des bornes he_sel1 et he_sel2 a partir de he_sel_deb et point
 * telles que he_sel1 <= he_sel2.
*/

void HeTextSel1Sel2 (int point)
{
    if (he_sel_deb < point)
         { he_sel1 = he_sel_deb; he_sel2 = point; }
    else { he_sel2 = he_sel_deb; he_sel1 = point; }
}


/*
 * Evenement KeyPress ; cf X11/keysymdef.h
*/

void HeTextKeyPress (He_node *hn, He_event *hev)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    int xsta  = hev->xev->xkey.state,
        shift = HE_BIT_ISSET (xsta, ShiftMask),
        alt1  = HE_BIT_ISSET (xsta, Mod1Mask);

    switch (hev->sym) {
            
    case XK_Tab :
    case XK_KP_Tab :
    case XK_ISO_Left_Tab :
        HeTextSelLine (hn);
        HeSetTextPoint (hn, HeGetTextLen(hn));
	break;
   
    case XK_Left : 
    case XK_KP_Left : 
    case XK_Right :
    case XK_KP_Right :
    case XK_Home : 
    case XK_KP_Home : 
    case XK_End :
    case XK_KP_End :
	if (shift && !HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) {
	    HeSetSelection (hn);
	    he_sel_deb = ob->point;
        }
        switch (hev->sym) {
          case XK_Left :
          case XK_KP_Left :  HeMoveTextPoint (hn, -1); break;
          case XK_Right :
          case XK_KP_Right : HeMoveTextPoint (hn, +1); break;
          case XK_Home :
          case XK_KP_Home :  HeSetTextPoint (hn, 0); break;
          case XK_End :
          case XK_KP_End :   HeSetTextPoint (hn, HeGetTextLen(hn)); break;
        }
        if (shift)
            HeTextSel1Sel2 (ob->point);
        else if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION))
            HeClearSelection () ;
	break;

    case XK_Delete :
    case XK_KP_Delete :
    case XK_BackSpace :
        if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) {
            HeSetTextPoint (hn, he_sel1);
	    HeTextDelete (hn, he_sel2 - he_sel1);
            HeClearSelection ();
        } else if (hev->sym == XK_BackSpace) {
            if (HeGetTextPoint(hn) != 0) {
                HeMoveTextPoint (hn, -1);
                HeMoveTextOffset(hn, -1);
                HeTextDelete (hn, 1);
            }
        } else {
	    HeTextDelete (hn, 1);
        }
        break;

    case XK_Insert :
    case XK_KP_Insert :
	HeSetTextInsertMode (hn, !HeGetTextInsertMode(hn));
        break;

    case XK_Return :
    case XK_KP_Enter :
    case XK_ISO_Enter :
	if (ob->notify_proc != NULL)
	    (ob->notify_proc)(hn);
	break;

    default :
    
        if (alt1) {
	    switch (hev->sym) {
              case 'x' : /* couper */ 
                HeTextCutClip (hn);
                return;
              case 'c' : /* copier */ 
                HeTextCopyClip (hn);
                return;
              case 'v' : /* coller */
                HeTextPasteClip (hn);
                return;
            }
        }
        
	if (hev->len > 0) {
        
            HeTextInput (hn, hev->str, hev->len, TRUE);

	    /* Filename completion si ob->point est a la fin */
	    if (HE_BIT_ISSET (ob->type, HE_TT_COMPLETION) &&
                ob->point == ob->str_len && ob->value && ob->value[0])
            {
                char s[256];
                if (HeCompletion (ob->value, s, 256) == 0) {
                    /* Lors du HeTextInsert il ne doit pas y avoir de selection
                       car l'effet de HeTextInsert en dpend. */
                    if (HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) 
                        HeClearSelection ();
		    he_sel_deb = he_sel1 = ob->point;
                    HeTextInsert (hn, s, strlen(s));
                    he_sel2 = ob->point;
                    HeSetSelection (hn);
                    HeSetTextPoint (hn, he_sel_deb);
                }
            }
            
	} /* else printf ("-- %s -- XK_%s   %d\n", 
		hev->str, XKeysymToString(hev->sym), hev->len); */
    }
}


/*
 * Les caracteres < 32 sont representes par "^A" pour 1, etc
*/

int HeTextSymLen (char c)
{
    return (c < 1 || c > 31) ? 1 : 2 ;
}

int HeTextSymStrLen (const char *s, int len)
{
    int res = 0, i;
    
    if (s == NULL) return 0;
    
    for (i = 0; s[i] && i < len; i++)
        res += HeTextSymLen (s[i]);
    return res;
}


/*
 * Recherche l'offset max de telle sorte que sym(fin de s) <= vis
 * Retour garanti dans [0,len]
*/

int HeTextSymMaxOffset (const char *s, int len, int vis)
{
    int i, tot = 0;
    
    if (s == NULL || len == 0 || vis == 0) return 0;
    
    for (i = len-1; i >= 0; i--) {
	tot += HeTextSymLen (s[i]);
        if (tot > vis) return i+1;
    }
    return 0;
}

/*
 * Detecte la position du caractere clique'
*/

int HeTextSymClic (He_node *hn, int x)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    int i, n = 0, x0, k, l;
    
    if (ob->value == NULL) return 0;
    
    x0 = HeTextCalcX0 (hn);
  
    for (i = ob->offset; ob->value[i] && n < ob->vis_len; i++) {
	l = HeTextSymLen (ob->value[i]); n += l; l *= he_font_width;
        k = l / 2; x0 += k;
        if (x <= x0) break;
        x0 += l - k;
    }
    return i;
}


/*
 * Calcule position de p dans la partie visible.
*/

int HeTextSymPos (He_node *hn, int p)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    int pos;
    
    if (p < ob->offset)  p = ob->offset;
    if (p > ob->str_len) p = ob->str_len;
    pos = HeTextSymStrLen (ob->value + ob->offset, p - ob->offset);
    if (pos > ob->vis_len) pos = ob->vis_len;
    return pos;
}


/*
 * Calcule deplacement realise par dx depuis le caractere numero p ; 
 * au retour, dx contient le reste du deplacement.
*/

int HeTextSymDepla (He_node *hn, int *dx, int p)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    int res = 0, l;
    
    if (p < 0) p = 0;
    if (p > ob->str_len) p = ob->str_len;
    
    if (*dx < 0) {
	while (p > 0) {
            l = HeTextSymLen (ob->value[--p]) * he_font_width;
            if (l > - *dx) break;
            *dx += l; res--;
        }    
    } else {
	while (p < ob->str_len) {
            l = HeTextSymLen (ob->value[p++]) * he_font_width;
            if (l > *dx) break;
            *dx -= l; res++;
        }
    }
    
    /* printf ("dx = %d  p = %d  res = %d\n", *dx, p, res); */
    return res;
}


/*
 * Affiche le texte avec les symboles
*/

void HeTextSymDraw (He_node *hn, int x, int y)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    char *buff, c;
    int i, n = 0;
    
    if (ob->value == NULL) return;
    
    buff = malloc (ob->vis_len + 32); /* Longueur max d'un sym <= 31 */
    if (buff == NULL) return;
    
    for (i = ob->offset; (c = ob->value[i]) && n < ob->vis_len; i++)
	if (c < 1 || c > 31) buff[n++] = c;
	else { buff[n++] = '^'; buff[n++] = '@' + c; }
    buff[n] = 0;
    
    HeDrawSubString(hn->win, he_gc_draw, he_normal_font,
	x, y, buff, 0, ob->vis_len);

    free (buff);
}


/*
 *  Teste si les bornes sont ok. 
*/
void HeTextBounds (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    int inf, sup;
    
    /* 0 <= ob->point <= ob->str_len */
    if (ob->point < 0) ob->point = 0;
    else if (ob->point > ob->str_len) ob->point = ob->str_len;
    
    /* limite gauche pour que ob->point soit visible */
    inf = HeTextSymMaxOffset (ob->value, ob->point, ob->vis_len);
    if (ob->offset < inf) ob->offset = inf;
    
    /* limite droite pour que ob->point soit visible */
    if (ob->offset > ob->point) ob->offset = ob->point;
    
    /* recherche offset sup pour fin de ligne a droite */
    sup = HeTextSymMaxOffset (ob->value, ob->str_len, ob->vis_len);
    if (ob->offset > sup) ob->offset = sup;
}


/*
 *  Teste nature du noeud ; renvoie TRUE si erreur. 
*/

int HeNotText (He_node *hn)
{
    if (hn == NULL || hn->meth != &he_meth_text) {
	HeError ("Text expected\n");
        return TRUE;
    }
    return FALSE;
}


