/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/panel.c - 28/06/1999
 *
 * 
*/

#include <helium.h>

/* Type et union du package 
*/
#define  OBJ_TYPE	Obj_panel
#define  OBJ_UNION	obj.panel

static He_method he_meth_panel;

#define HE_X_ALI  4
#define HE_Y_ALI  3
#define HE_X_GAP  6
#define HE_Y_GAP  5


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 *  Cration d'un panel.
 *  Renvoie le noeud ou NULL.
*/

He_node *HeCreatePanel (He_node *owner)
{
    He_node *hn;
    OBJ_TYPE *ob;

    HePanelAutoInit();

    /* On cree le noeud */
    hn = HeBaseCreate (owner, &he_meth_panel);
    if (hn == NULL) return NULL;

    /* On cree l'objet */
    ob = hn->OBJ_UNION = malloc (sizeof (OBJ_TYPE));
    if (ob == NULL) {
	HeError ("HeCreatePanel: malloc error\n");
	return NULL;
    }

    /* Init des champs du noeud */
    hn->xm = hn->owner->xm;
    hn->ym = hn->owner->ym;
    
    /* init des champs de l'objet */
    ob->layout = HE_HORIZONTAL;
    ob->x_ali  = HE_X_ALI;
    ob->y_ali  = HE_Y_ALI;
    ob->x_gap  = HE_X_GAP;
    ob->y_gap  = HE_Y_GAP;
    ob->ref_item = 0;
    ob->pointed = 0;
    ob->kbd_proc = NULL;

    /* Creation du Window avec Xlib */
    hn->win_border = 0;
    hn->win = XCreateSimpleWindow (he_display, 
	hn->owner->win,
	hn->xb, hn->yb, hn->xm, hn->ym, 
	hn->win_border,
	he_black,
	he_col_bg );

    /* Proprit du window */
    XSetStandardProperties (he_display, hn->win, "", "",
	0, he_argv, he_argc, NULL);
    HeEventMask (hn->win);
    HeWinListInsert (hn);

    /* Update diffr */
    HeUpdateMark (hn, HE_UPD_DRAW);

    /* Fin succes */
    return hn;
}


/*
 * Set
*/

void HeSetPanelLayout (He_node *hn, int layout)
{
    if (HeNotPanel(hn)) return;   
    hn->OBJ_UNION->layout = layout;
}

void HeSetPanelXAlign (He_node *hn, int x_ali)
{
    if (HeNotPanel(hn)) return;   
    hn->OBJ_UNION->x_ali = x_ali;
}

void HeSetPanelYAlign (He_node *hn, int y_ali)
{
    if (HeNotPanel(hn)) return;   
    hn->OBJ_UNION->y_ali = y_ali;
}

void HeSetPanelXGap (He_node *hn, int x_gap)
{
    if (HeNotPanel(hn)) return;   
    hn->OBJ_UNION->x_gap = x_gap;
}

void HeSetPanelYGap (He_node *hn, int y_gap)
{
    if (HeNotPanel(hn)) return;   
    hn->OBJ_UNION->y_gap = y_gap;
}

void HeSetPanelRef (He_node *hn, He_node *ref)
{
    if (HeNotPanel(hn)) return;   
    hn->OBJ_UNION->ref_item = (ref == NULL) ? 0 : ref->id;
}

void HeSetPanelKbdProc (He_node *hn, He_kbd_proc proc)
{
    if (HeNotPanel(hn)) return; 
    hn->OBJ_UNION->kbd_proc = proc;
}


/*
 * Get
*/

int HeGetPanelLayout (He_node *hn)
{
    if (HeNotPanel(hn)) return HE_HORIZONTAL;
    return hn->OBJ_UNION->layout;
}

int HeGetPanelXAlign (He_node *hn)
{
    if (HeNotPanel(hn)) return 0;
    return hn->OBJ_UNION->x_ali;
}

int HeGetPanelYAlign (He_node *hn)
{
    if (HeNotPanel(hn)) return 0;
    return hn->OBJ_UNION->y_ali;
}

int HeGetPanelXGap (He_node *hn)
{
    if (HeNotPanel(hn)) return 0;
    return hn->OBJ_UNION->x_gap;
}

int HeGetPanelYGap (He_node *hn)
{
    if (HeNotPanel(hn)) return 0;
    return hn->OBJ_UNION->y_gap;
}

He_node *HeGetPanelRef (He_node *hn)
{
    if (HeNotPanel(hn)) return NULL;
    return HeGetNodeFromId (hn->OBJ_UNION->ref_item);
}

He_kbd_proc HeGetPanelKbdProc (He_node *hn)
{
    if (HeNotPanel(hn)) return NULL; 
    return hn->OBJ_UNION->kbd_proc;
}


/*
 *  Teste nature du noeud ; renvoie FALSE si erreur ; silencieux.
*/

int HeIsPanel (He_node *hn)
{
    return (hn == NULL || hn->meth != &he_meth_panel) ? FALSE : TRUE;
}


/*-------------------- M E T H O D E S - P R I V E E S ------------------------
 *
 * On ne teste pas si hn == NULL.
*/

/*
 * Initialisation du module ; appel par HeCreatePanel().
*/

void HePanelAutoInit ()
{
    static int did_init = FALSE;
    
    if (did_init) return;
    
    HeMethodDefault (&he_meth_panel, HE_NAT_PANEL);
    he_meth_panel.props   = HE_PM_WIN_OWNER;
    he_meth_panel.destroy = HePanelDestroy;
    he_meth_panel.fit     = HePanelFit;
    he_meth_panel.update  = HePanelUpdate;
    he_meth_panel.event   = HePanelEvent;
    
    did_init = TRUE;
}


/*
 * Destruction d'un panel ; appele' par HeDestroyRec().
*/

void HePanelDestroy (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;

    HeWinListRemove (hn);
    XDestroyWindow (he_display, hn->win);

    free (ob);
}


/*
 * Ajuste la taille du panel en fonction de son contenu.
*/

void HePanelFit (He_node *hn)
{
    int xmax, ymax;

    if (HeNotPanel(hn)) return;
    
    HeBoundingBox (hn, &xmax, &ymax);
    HeSetWidth (hn, xmax + hn->OBJ_UNION->x_ali);
    HeSetHeight(hn, ymax + hn->OBJ_UNION->y_ali);
}


/*
 * M-A-J geometrie et affichage ; appele' par HeUpdateFlush().
*/

void HePanelUpdate (He_node *hn, He_upd_info *ui)
{
    HeUpdateWinGeom (hn, ui);

    if (HE_BIT_ISSET (ui->mask, HE_UPD_BORDER)) {
	XSetWindowBorderWidth (he_display, hn->win, hn->win_border);
    }

    if (HE_BIT_ISSET (ui->mask, HE_UPD_DRAW)) {
	if (HE_BIT_ISSET (hn->status, HE_STA_SHOW))
	     XMapWindow   (he_display, hn->win);
	else XUnmapWindow (he_display, hn->win);
    }
}


/*
 * Methode pour gerer les evenements X.
 * Renvoie -1 si erreur.
 *
 * Appele' par HeEventDispatch()
*/

int HePanelEvent (He_node *hn, He_event *hev)
{
    OBJ_TYPE *ob = hn->OBJ_UNION ;

#ifdef DEBUG
    if (hev->xev->type != MotionNotify)
        printf ("panel %s\n", HeGetXEventName (hev->type)); 
#endif

    return HeEventToGadgets (hn, hev, ob->kbd_proc, &ob->pointed);
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * ATTENTION ici hn est le panel_item, hn->owner est le panel.
 * Positionne le panel_item par rapport au layout.
 *
 * Appele' par HePanel**Create()
*/
void HePosPanelItem (He_node *hn)
{
    OBJ_TYPE *ob_pan;
    He_node *hn_ref;

    /* Si le owner n'est pas un panel, alors sortie silencieuse */
    if (hn->owner->meth != &he_meth_panel) return;
    
    ob_pan = hn->owner->OBJ_UNION ;
    hn_ref = HeGetNodeFromId (ob_pan->ref_item);
    
    if (hn_ref == NULL) {
	hn->xb = ob_pan->x_ali; 
	hn->yb = ob_pan->y_ali;
    } else if (ob_pan->layout == HE_HORIZONTAL) {
	hn->xb = hn_ref->xb + hn_ref->xm + ob_pan->x_gap;
	hn->yb = hn_ref->yb;
    } else if (ob_pan->layout == HE_VERTICAL) {
	hn->xb = hn_ref->xb;
	hn->yb = hn_ref->yb + hn_ref->ym + ob_pan->y_gap;
    } else if (ob_pan->layout == HE_LINE_FEED) {
	hn->xb = ob_pan->x_ali;
	hn->yb = hn_ref->yb + hn_ref->ym + ob_pan->y_gap;
	ob_pan->layout = HE_HORIZONTAL;
    }
    /* Le nouvel item de reference */
    ob_pan->ref_item = hn->id;
}


/*
 *  Teste nature du noeud ; renvoie TRUE si erreur. 
*/

int HeNotPanel (He_node *hn)
{
    if (hn == NULL || hn->meth != &he_meth_panel) {
	HeError ("Panel expected\n");
        return TRUE;
    }
    return FALSE;
}


