/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/wm.c - 10/10/2000
 *
 * Fonctions de paramtrage avec le Window Manager.
*/

#include <helium.h>


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Place un Hint pour que les fenetres s'iconifient lorsque he_main_frame
 * est iconifiee.
 * 
 * xprop : WM_CLIENT_LEADER(WINDOW): window id # 0x....
*/

void HeWMGroup (He_node *hn)
{
    /* On rappellera plus tard */
    if (he_main_frame == NULL) return;
    
    /* Y a-t'il une fonction Xlib qui fait cela ? */
    HeAtomSet32 (hn->win, he_atom_leader, XA_WINDOW, (int) he_main_frame->win);
}


/*
 * Place un Hint pour que le WM envoie les evt FocusIn/FocusOut.
 * 
 * xprop : WM_HINTS(WM_HINTS):
 *		Client accepts input or input focus: True
*/

void HeWMFocus (He_node *hn)
{
    /* On rcupre les wmhints existants pour ne pas les craser */
    XWMHints *p_hints = XGetWMHints (he_display, hn->win);
    if (p_hints != NULL) {
	p_hints->input = TRUE;
	HE_BIT_SET (p_hints->flags, InputHint);
	XSetWMHints (he_display, hn->win, p_hints);
	XFree (p_hints);
    } else {
        XWMHints hints;
	hints.input = TRUE;
	hints.flags = InputHint;
	XSetWMHints (he_display, hn->win, &hints);
    }
}


/*
 * Protege d'un "Quit" sauvage du WM.
 * 
 * xprop : WM_PROTOCOLS(ATOM): protocols  WM_DELETE_WINDOW
*/

void HeWMProtocols (He_node *hn)
{
    Atom liste[1];
    liste[0] = he_atom_delwin;
    XSetWMProtocols (he_display, hn->win, liste, 1);
}


/*
 * Donne les noms de ressource et de classe.
 * 
 * xprop : WM_CLASS(STRING) = "ressource", "classe"
*/

void HeWMClass (He_node *hn)
{
    XClassHint class_hints;
    class_hints.res_name = he_res_name;
    class_hints.res_class = he_res_class;
    XSetClassHint (he_display, hn->win, &class_hints);
}


/*
 * Place dans les "Hints" la position de la fenetre
 * Le Window Manager s'en sert lorsqu'elle est remappe'e
 *
 * xprop : WM_NORMAL_HINTS(WM_SIZE_HINTS):
 *              user specified location: x, y
*/

void HeWMHintPos (He_node *hn)
{
    XSizeHints hints;
    if (!XGetNormalHints (he_display, hn->win, &hints)) hints.flags = 0;
    hints.x = hn->xb;  hints.y = hn->yb;
    HE_BIT_SET (hints.flags, USPosition);
    XSetNormalHints (he_display, hn->win, &hints);
}


/*
 * Place dans les "Hints" la taille de la fenetre
 * Le Window Manager s'en sert lorsqu'elle est remappe'e
 *
 * xprop : WM_NORMAL_HINTS(WM_SIZE_HINTS):
 *              user specified size: w by h
*/

void HeWMHintSize (He_node *hn)
{
    XSizeHints hints;
    if (!XGetNormalHints (he_display, hn->win, &hints)) hints.flags = 0;
    hints.width = hn->xm;  hints.height = hn->ym;
    HE_BIT_SET (hints.flags, USSize);
    XSetNormalHints (he_display, hn->win, &hints);
}


/*
 * Met le flag override_redirect  True.
 * Le Window Manager ne filtre plus les requtes de gomtrie.
*/

void HeWMOverrideRedirect (He_node *hn)
{
    XSetWindowAttributes attr;
    attr.override_redirect = True;
    XChangeWindowAttributes (he_display, hn->win, CWOverrideRedirect, &attr);
}


/*
 * Met le flag save_under  True. Demande que l'affichage sous le Window soit
 * mmoris et restaur automatiquement.
*/

void HeWMSaveUnder (He_node *hn)
{
    XSetWindowAttributes attr;
    attr.save_under = True;
    XChangeWindowAttributes (he_display, hn->win, CWSaveUnder, &attr);
}

